import logging

from domiknows.solver.ilpBooleanMethods import ilpBooleanProcessor 
from domiknows.solver.ilpConfig import ilpConfig 

from gurobipy import Var, GRB, LinExpr

USE_De_Morgan = False # For orVar nandVar methods 

# Creates ILP constraints for logical methods based on provided arguments.
#
# The method's arguments originate either directly from the candidates (dataNode) 
#                                  or are the result of the nested logical constraints evaluation.
#
# The method's arguments can be:
#   - ILP variable, 
#   - number (0 or 1) representing True or False value,
#   - None representing lack of information (when the candidate is missing in the dataNode).

class gurobiILPBooleanProcessor(ilpBooleanProcessor):
    
    def __init__(self, _ildConfig = ilpConfig) -> None:
        super().__init__()
                
        self.grad = False
        
        self.myLogger = logging.getLogger(ilpConfig['log_name'])
        self.ifLog =  ilpConfig['ifLog']
                
    def __varIsNumber(self, var):
        return not isinstance(var, Var)
    
    def preprocessLogicalMethodVar(self, var, logicMethodName, varNameConnector, minN = 2):
        # -- Check types of variables in var - gather information about them
        varsInfo = {}
        varsInfo['N'] = len(var)                # Number of all variables
        varsInfo['iLPVars'] = []                # ILP variables
        varsInfo['No_of_ilp'] = 0               # Number of IPL variables
        varsInfo['varsNames'] = []              # Names of variables
        varsInfo['varName'] = ""                # Name of the new ILP variable if created
        varsInfo['numberMul'] = 1               # Multiplication of all numbers if present
        varsInfo['numberSum'] = 0               # Summation of all numbers if present
        varsInfo['varSumLinExpr'] = LinExpr()   # Summation of all ILP variable
        varsInfo['varSumLinExprStr'] = ""       # String representation of the summation of all ILP variable
        
        for i, currentVar in enumerate(var):
            if self.__varIsNumber(currentVar):
                varsInfo['varsNames'].append(currentVar)
                varsInfo['numberMul'] *= currentVar
                varsInfo['numberSum'] += currentVar
            else:
                try:
                    varsInfo['varsNames'].append(currentVar.VarName)
                except AttributeError:
                    pass
                
                varsInfo['iLPVars'].append(currentVar)
                varsInfo['No_of_ilp'] += 1

                varsInfo['varSumLinExpr'].addTerms(1.0, currentVar)

                varsInfo['varName'] += varNameConnector
                varsInfo['varName'] += "_%s_" % (currentVar.VarName)
    
        if varsInfo['varSumLinExpr'].size() > 0:
            varsInfo['varSumLinExprStr'] = str(varsInfo['varSumLinExpr']) 
            #varsInfo['varSumLinExprStr'] = varsInfo['varSumLinExprStr'][varsInfo['varSumLinExprStr'].index(':') + 1 : varsInfo['varSumLinExprStr'].index('>')]
          
        if len(varsInfo['varName']) > 0:   
            varsInfo['varName'] = varsInfo['varName'][:-1] # Remove last '_'
            varsInfo['varName'] = '{:.200}'.format(varsInfo['varName'])
            varsInfo['varName'] = varsInfo['varName'][:254] # Limit size of the new ILP variable name
        
        if self.ifLog: self.myLogger.debug("%s called with: %s"%(logicMethodName, varsInfo['varsNames']))
        
        if varsInfo['N'] < minN: # Variables number less than min 
            raise Exception("%s has no enough variable - %i, required %i"%(logicMethodName,varsInfo['N'],minN))
            
        return varsInfo
    
    def notVar(self, m, var, onlyConstrains = False):
        logicMethodName = "NOT"
        
        # -- Consider None
        varFixed = []  
        if var is None:
            varFixed.append(0) # when None
        else:
            varFixed.append(var)
        # --
        
        varsInfo = self.preprocessLogicalMethodVar(varFixed, logicMethodName, "not", minN=1)
        
        if varsInfo['N'] > 1: # More than 1 variable
            raise Exception("%s has %i variables, accepts only 1"%(logicMethodName,varsInfo['N']))
        
        # -- Only constructing constrains forcing NOT to be True 
        if onlyConstrains:
            if varsInfo['No_of_ilp'] == 0: # Called with a number
                if varsInfo['numberSum'] == 0: # number is 0
                    # Applying not results in True
                    return 
                else: # number is 1
                    # Applying not results in False -> model is infeasible -> exception
                    raise Exception("ILP model is infeasible - %s is called with value %i, and the result of applying %s is False"%(logicMethodName,1,logicMethodName))
            else:
                # -- Create constraint as there is an ILP variable
                m.addConstr(varsInfo['iLPVars'][0] == 0, name='Not:') # ILP variable has to be 0 so applying not will result in True
                if self.ifLog: self.myLogger.debug("%s created constraint only: not %s == %i"%(logicMethodName,varsInfo['varsNames'][0],0))
    
                return
        else:
            # --- Creating ILP variable representing value of NOT build of provided method argument
            
            if varsInfo['No_of_ilp'] == 0: # Called with a number
                if varsInfo['numberSum'] == 0:
                    # Applying not results in True
                    if self.ifLog: self.myLogger.debug("%s returns: %i"%(logicMethodName,1))
                    return 1
                else:
                    # Applying not results in False
                    if self.ifLog: self.myLogger.debug("%s returns: %i"%(logicMethodName,0))
                    return 0
            else: 
                # -- One ILP variable
                
                # Create new ILP variable 
                varNOT = m.addVar(vtype=GRB.BINARY, name=varsInfo['varName'])
                if m: m.update()
        
                # Adding ILP constraint
                m.addConstr(varNOT + varsInfo['iLPVars'][0] == 1, name='Not:') 
                if self.ifLog: self.myLogger.debug("%s created constraint: %s + %s == %i "%(logicMethodName,varNOT.VarName,varsInfo['varsNames'][0],1))
        
                if self.ifLog: self.myLogger.debug("%s returns: %s"%(logicMethodName,varsInfo['varName']))
                return varNOT
            
    def and2Var(self, m, var1, var2, onlyConstrains = False):
        return self.andVar(m, (var1, var2), onlyConstrains=onlyConstrains)
    
    def andVar(self, m, *var, onlyConstrains = False):
        logicMethodName = "AND"
        
        # -- Consider None
        varFixed = []  
        for v in var:
            if v is None:
                varFixed.append(1) # when None
            else:
                varFixed.append(v)
        # --
        
        varsInfo = self.preprocessLogicalMethodVar(varFixed, logicMethodName, "and")
       
        # -- Only constructing constrains forcing AND to be True 
        if onlyConstrains:    
            if varsInfo['numberMul'] == 0: # Vars numbers multiply to 0 - at least one zero present
                # Applying and results in False -> model is infeasible -> exception
                raise Exception("ILP model is infeasible - %s is called with value %i, and the result of applying %s is False"%(logicMethodName,0,logicMethodName))
            elif varsInfo['No_of_ilp'] == 0: # No ILP variables
                # Applying and to all 1 (multiply is 1) results in True
                if self.ifLog: self.myLogger.debug("%s returns: %i"%(logicMethodName,1))
                return
            elif varsInfo['No_of_ilp'] == 1: # if only one ILP variable - rest have to be 1 as there were not zeros
                # Adding ILP constraint
                m.addConstr(varsInfo['iLPVars'][0] == 1, name='Not:') # ILP variable has to be 1 so applying and will result in True
                return
            else:
                # -- Create constraint as there are at least two ILP variables and all numbers, if present are 1
                m.addConstr(varsInfo['No_of_ilp'] - varsInfo['varSumLinExpr'] <= 0, name='And:') #  varSumLinExpr >= N
                if self.ifLog: self.myLogger.debug("%s created constraint only: and %s > %i"%(logicMethodName,varsInfo['varSumLinExprStr'],varsInfo['No_of_ilp']))
                return
        else:  
            # -- If creating ILP variable representing value of AND build of provided method arguments
            
            if varsInfo['numberMul'] == 0: # Vars numbers multiply to 0 - at least one 0 present
                # Applying and results in False
                if self.ifLog: self.myLogger.debug("%s has zero, returning 0 without creating additional constraint"%(logicMethodName))
                return 0
            elif varsInfo['No_of_ilp'] == 0: # No ILP variables
                # Applying and results in True
                if self.ifLog: self.myLogger.debug("%s has no ILP variable, returning %i without creating additional constraint"%(logicMethodName, varsInfo['numberMul']))
                return 1
            elif varsInfo['No_of_ilp'] == 1: # Only single ILP variable; rest is 1 here
                # Result of and is the value of the single ILP variable
                if self.ifLog: self.myLogger.debug("%s has ones and only single variable: %s, it is returned"%(logicMethodName,varsInfo['iLPVars'][0]))
                return varsInfo['iLPVars'][0]
            else:      
                # -- More than one ILP variable and the rest is 1 
                
                # Create new variable
                varAND = m.addVar(vtype=GRB.BINARY, name=varsInfo['varName'])
                if m: m.update()
        
                # Build ILP constraints 
                for currentVar in varsInfo['iLPVars']:
                    m.addConstr(varAND - currentVar <= 0, name='And:') # varAND <= currentVar
                
                # Adding ILP constraint
                m.addConstr(varsInfo['varSumLinExpr'] - varAND <= varsInfo['No_of_ilp'] - 1, name='And:') #  varSumLinExpr <= varAND + N - 1
    
                if self.ifLog: self.myLogger.debug("%s returns: %s"%(logicMethodName,varsInfo['varName']))
                return varAND
    
    def or2Var(self, m, var1, var2, onlyConstrains = False):
        return self.orVar(m, (var1, var2), onlyConstrains = onlyConstrains)
    
    def orVar(self, m, *var, onlyConstrains = False):
        if USE_De_Morgan:
            notVar = []
            for v in var:
                notVar.append(self.notVar(m, v))
            
            return self.notVar(m, self.andVar(m, *notVar), onlyConstrains=onlyConstrains) #  Negation of the conjunction of the negations
        
        logicMethodName = "OR"
        
        # -- Consider None
        varFixed = []  
        for v in var:
            if v is None:
                varFixed.append(0) # when None
            else:
                varFixed.append(v)
        # --
        
        varsInfo = self.preprocessLogicalMethodVar(varFixed, logicMethodName, "or")
        
        # If only constructing constrains forcing OR to be True 
        if onlyConstrains:
            if varsInfo['numberSum'] > 0: # Vars numbers sum is larger then 0 - at least one is present
                # Applying or results in True
                if self.ifLog: self.myLogger.debug("%s has ones, returning without creating constraint"%(logicMethodName))
                return
            elif varsInfo['No_of_ilp'] == 0: # No ILP variables
                # Applying or results in False -> model is infeasible -> exception
                raise Exception("ILP model is infeasible - %s is called with values %i, and the result of applying %s is False"%(logicMethodName,0,logicMethodName))
            elif varsInfo['No_of_ilp'] == 1: # Only one ILP variable and the rest are zeros
                # Adding ILP constraint
                m.addConstr(varsInfo['iLPVars'][0] >= 1, name='Or:') # ILP variable has to be 1 so applying or will result in True
                if self.ifLog: self.myLogger.debug("%s created constraint only: %s >= %i"%(logicMethodName,varsInfo['iLPVars'][0],1))
                return
            else:
                # -- Create constraint as there are at least two ILP variables and all numbers, if present, are 0         
                m.addConstr(varsInfo['varSumLinExpr'] >= 1, name='Or:') # varSumLinExpr >= 1
                if self.ifLog: self.myLogger.debug("%s created constraint only: %s >= %i"%(logicMethodName,varsInfo['varSumLinExprStr'],1))
                return
        else:
            # ------- Creating ILP variable representing value of OR build of provided method arguments
            
            if varsInfo['numberSum'] > 0: #  Vars numbers sum is larger then 0 - at least one present
                # Applying or results in True
                if self.ifLog: self.myLogger.debug("%s has ones, returning 1 without creating additional constraint"%(logicMethodName))
                return 1
            elif varsInfo['No_of_ilp'] == 0: # No ILP variables
                if self.ifLog: self.myLogger.debug("%s has no ILP variable, returning %i without creating additional constraint"%(logicMethodName, varsInfo['numberSum']))
                return varsInfo['numberSum']
            elif varsInfo['No_of_ilp'] == 1: # Only single ILP variable; rest has to be zeros: see first if above
                if self.ifLog: self.myLogger.debug("%s has zeros and only single variable: %s, it is returned"%(logicMethodName,varsInfo['iLPVars'][0]))
                return varsInfo['iLPVars'][0]
            else:
                # -- More than one ILP variable and the rest is 0 
                
                # Create new variable
                varOR = m.addVar(vtype=GRB.BINARY, name=varsInfo['varName'])
                if m: m.update()
        
                # Build constrains
                for currentVar in varsInfo['iLPVars']:
                    m.addConstr(currentVar - varOR <= 0, name='Or:') # currentVar <= varOR
                    if self.ifLog: self.myLogger.debug("%s created constraint: %s - %s <= %i"%(logicMethodName,currentVar.VarName,varsInfo['varName'],0))
        
                m.addConstr(varsInfo['varSumLinExpr'] - varOR >= 0, name='Or:') # varSumLinExpr >= varOR
                if self.ifLog: self.myLogger.debug("%s created constraint: %s - %s >= %i"%(logicMethodName,varsInfo['varSumLinExprStr'],varsInfo['varName'],1-1))
        
                if self.ifLog: self.myLogger.debug("%s returns new variable: %s"%(logicMethodName,varsInfo['varName']))
                return varOR
            
    def nand2Var(self, m, var1, var2, onlyConstrains = False):
        return self.nandVar(m, (var1, var2), onlyConstrains = onlyConstrains)
    
    def nandVar(self, m, *var, onlyConstrains = False):
        logicMethodName = "NAND"
       
        if USE_De_Morgan:
            notVar = []
            for v in var:
                notVar.append(self.notVar(m, v))
            
            return self.andVar(m, *notVar, onlyConstrains=onlyConstrains) # Negation of their conjunction 
        
        # -- Consider None
        varFixed = []  
        for v in var:
            if v is None:
                varFixed.append(0) # when None
            else:
                varFixed.append(v)
        
        var = varFixed
        # --
        
        varsInfo = self.preprocessLogicalMethodVar(varFixed, logicMethodName, "or")
                
        # If only constructing constrains forcing NAND to be true 
        if onlyConstrains:
            varSumLinExpr = LinExpr()
            for currentVar in var:
                varSumLinExpr.addTerms(1.0, currentVar)
        
            m.addConstr(varSumLinExpr <= varsInfo["N"] - 1, name='Nand:')     
            if self.ifLog: self.myLogger.debug("NAND created constraint only: %s <= %i"%(varsInfo["varSumLinExprStr"], varsInfo["N"]-1))
                  
            return
        
        # ------- If creating variables representing value of NAND build of provided variables
        varNAND = m.addVar(vtype=GRB.BINARY, name=varsInfo['varName'])
        for currentVar in var:
            m.addConstr(self.notVar(m, varNAND) <= currentVar, name='Nand:')

        m.addConstr(varsInfo["varSumLinExpr"]<= self.notVar(m, varNAND) + varsInfo["N"] - 1, name='Nand:')
    
        return varNAND
    
    def nor2Var(self, m, var1, var2, onlyConstrains = False):
        return self.norVar(m, (var1, var2), onlyConstrains = onlyConstrains)
    
    def norVar(self, m, *var, onlyConstrains = False):
        
        return self.notVar(m, self.orVar(m, var), onlyConstrains=onlyConstrains) # Negation of the disjunction
    
        #---------------------- No used anymore
        
        # -- Consider None
        varFixed = []  
        for v in var:
            if v is None:
                varFixed.append(1)
            else:
                varFixed.append(v)
        
        var = varFixed
        # --
        
        N = len(var)
        
        if N <= 1:
            return None
        
        if onlyConstrains:
            varSumLinExpr = LinExpr()
            for currentVar in var:
                varSumLinExpr.addTerms(1.0, currentVar)
        
            m.addConstr(varSumLinExpr <= 0)
            return
        
        _norVarName = "nor"
        for currentVar in var:
            _norVarName += "_%s"%(currentVar)
           
        _norVarName = _norVarName [:254]
        varNOR = m.addVar(vtype=GRB.BINARY, name=_norVarName)
        for currentVar in var:
            m.addConstr(currentVar <= self.notVar(m, varNOR))
        
        varSumLinExpr = LinExpr()
        for currentVar in var:
            varSumLinExpr.addTerms(1.0, currentVar)
            
            m.addConstr(varSumLinExpr >= self.notVar(m, varNOR))
    
        return varNOR
    
    def xorVar(self, m, var1, var2, onlyConstrains = False):
        var = (var1, var2)
        
        # Conjunction of the disjunction and the negation of the conjunction
        return self.andVar(m, self.orVar(m, var), self.notVar(m, self.andVar(m, var)), onlyConstrains=onlyConstrains) 
                
        #---------------------- No used anymore

        # -- Consider None
        if var1 is None:
            var1 = 1
            
        if var2 is None:
            var2 = 1
        # --

        # Get names of ILP variables
        var1Name = var1
        var2Name = var2
        if not self.__varIsNumber(var1):
            var1Name = var1.VarName
        if not self.__varIsNumber(var2):
            var2Name = var2.VarName
            
        if onlyConstrains:
            m.addConstr(var1 + var2 <= 1)
            m.addConstr(var1 + var2 >= 1)
            if self.ifLog: self.myLogger.debug("IF created constraint only: %s <= %s"%(var1Name, var2Name))

            return
        
        varXORName = "xor_%s_%s"%(var1, var2)
        varXORName = varXORName[:254]
        varXOR = m.addVar(vtype=GRB.BINARY, name=varXORName)
            
        m.addConstr(var1 + var2 + varXOR <= 2)
        m.addConstr(-var1 - var2 + varXOR <= 0)
        m.addConstr(var1 - var2 + varXOR >= 0)
        m.addConstr(-var1 + var2 + varXOR >= 0)
            
        return varXOR
    
    def ifVar(self, m, var1, var2, onlyConstrains = False):
        logicMethodName = "IF"

        # -- Consider None
        hasNone = False
        if var1 is None: # antecedent 
            antecedent = 1 # when None
            self.myLogger.info("%s called with antecedent equals None"%(logicMethodName))
            self.myLogger.info("%s called with consequent equals %s"%(logicMethodName,var2))

            hasNone = True
        else:
            antecedent = var1

        if var2 is None: # consequent
            consequent = 0 # when None
            if not hasNone: # not yet
                self.myLogger.info("%s called with antecedent equals %s"%(logicMethodName,var1))
                self.myLogger.info("%s called with consequent equals None"%(logicMethodName))
                hasNone = True
        else:
            consequent = var2
        # --
    
        varsInfo = self.preprocessLogicalMethodVar((antecedent,consequent), logicMethodName, "if",  minN=2)
        
        if varsInfo['N'] > 2: # More than 2 variable
            raise Exception("%s has %i variables, accepts only 2"%(logicMethodName,varsInfo['N']))
            
        if onlyConstrains:
            if varsInfo['No_of_ilp'] == 0: # No ILP variables
                if (not antecedent): 
                    # Applying if results in True
                    if self.ifLog: self.myLogger.debug("%s is True - antecedent is False"%(logicMethodName))
                    if hasNone: self.myLogger.info("%s is True - antecedent is False"%(logicMethodName))
                    return 
                elif consequent: # antecedent is True
                    # Applying if results in True
                    if self.ifLog: self.myLogger.debug("%s is True - antecedent and consequent are True"%(logicMethodName))
                    if hasNone: self.myLogger.info("%s is True - antecedent and consequent are True"%(logicMethodName))
                    return 
                else: # antecedent and not consequent
                    raise Exception("ILP model is infeasible - %s is called with the antecedent True and the consequent False - the result of applying %s is False"
                                    %(logicMethodName,logicMethodName))
            elif self.__varIsNumber(antecedent): # antecedent is boolean and consequent is the ILP variable
                if not antecedent:
                    # Applying if results in True
                    if self.ifLog: self.myLogger.debug("%s is True - antecedent is False"%(logicMethodName))
                    if hasNone: self.myLogger.info("%s is True - antecedent is False"%(logicMethodName))
                    return 
                else: # antecedent is True
                    # Adding ILP constraint
                    m.addConstr(consequent >= 1, name='If:') # var2 >= 0 - consequent has to be True
                    if self.ifLog: self.myLogger.debug("%s antecedent is True - created constraint: %s >= %i"%(logicMethodName,consequent,1))
                    if hasNone: self.myLogger.info("%s antecedent is True - created constraint: %s >= %i"%(logicMethodName,consequent,1))
                    return
            elif  self.__varIsNumber(consequent): # consequent is boolean and the antecedent is the ILP variable
                if consequent:
                    # Applying if results in True
                    if self.ifLog: self.myLogger.debug("%s is True - consequent is True"%(logicMethodName))
                    if hasNone: self.myLogger.info("%s is True - consequent is True"%(logicMethodName))
                    return 
                else: # consequent is False
                    # Adding ILP constraint - antecedent ILP variable has to be False too
                    m.addConstr(antecedent <= 0, name='If:')
                    if self.ifLog: self.myLogger.debug("%s consequent is False - created constraint: %s <= %i"%(logicMethodName,antecedent,0))
                    if hasNone:  self.myLogger.info("%s consequent is False - created constraint: %s <= %i"%(logicMethodName,antecedent,0))
                    return
                        
            # Create constraint as there are two ILP variables         

            # Only constructing constrains forcing IF to be true 
            m.addConstr(antecedent - consequent <= 0, name='If:') # var1 <= var2
            if self.ifLog: self.myLogger.debug("%s created constraint only: %s <= %s"%(logicMethodName,varsInfo['varsNames'][0],varsInfo['varsNames'][1]))
            
            return
        else:
            if varsInfo['No_of_ilp'] == 0: # No ILP variable
                if not antecedent:
                    if self.ifLog: self.myLogger.debug("%s is True - antecedent is False - returning %i"%(logicMethodName,1))
                    if hasNone: self.myLogger.info("%s is True - antecedent is False - returning %i"%(logicMethodName,1))
                    return 1
                elif consequent: # antecedent is True
                    if self.ifLog: self.myLogger.debug("%s is True - antecedent and consequent are True - returning %i"%(logicMethodName,1))
                    if hasNone: self.myLogger.info("%s is True - antecedent and consequent are True - returning %i"%(logicMethodName,1))
                    return 1
                else: # antecedent and not consequent
                    if self.ifLog: self.myLogger.debug("%s is False - is called with the antecedent True and the consequent False - returning %i"%(logicMethodName,0))
                    if hasNone: self.myLogger.info("%s is False - is called with the antecedent True and the consequent False - returning %i"%(logicMethodName,0))
                    return 0
            elif self.__varIsNumber(antecedent):
                if not antecedent:
                    if self.ifLog: self.myLogger.debug("%s is True - antecedent is False - returning %i"%(logicMethodName,1))
                    if hasNone: self.myLogger.info("%s is True - antecedent is False - returning %i"%(logicMethodName,1))
                    return 1
                else: #antecedent
                    if self.ifLog: self.myLogger.debug("%s returns: %s - antecedent is True"%(logicMethodName,consequent))
                    if hasNone: self.myLogger.info("%s returns: %s - antecedent is True"%(logicMethodName,consequent))
                    return consequent
            elif  self.__varIsNumber(consequent):
                if consequent:
                    if self.ifLog: self.myLogger.debug("%s is True - consequent is True - returning %i"%(logicMethodName,1))
                    if hasNone: self.myLogger.info("%s is True - consequent is True - returning %i"%(logicMethodName,1))
                    return 1
                else: #not consequent
                    notAntecedent = self.notVar(m, antecedent)
                    if self.ifLog: self.myLogger.debug("%s returns: %s - consequent is False"%(logicMethodName,notAntecedent.VarName))
                    if hasNone: self.myLogger.info("%s returns: %s - consequent is False"%(logicMethodName,notAntecedent.VarName))
                    return notAntecedent
       
            # Create new variable
            varIF = m.addVar(vtype=GRB.BINARY, name=varsInfo["varName"])
                
            # Build constrains
            m.addConstr(1 - antecedent <= varIF, name='If:')                # 1 - var1 <= varIF
            m.addConstr(consequent <= varIF, name='If:')                    # var2 <= varIF
            m.addConstr(1 - antecedent + consequent >= varIF, name='If:')   # 1- var1 + var2 >= varIF
                
            m.update()
            
            if self.ifLog: self.myLogger.debug("IF returns : %s"%(varsInfo["varName"]))
            return varIF
           
    def epqVar(self, m, var1, var2, onlyConstrains = False):
        #if self.ifLog: self.myLogger.debug("EQ called with : %s"%(var1,var2))

        # -- Consider None
        if var1 is None:
            var1 = 0
            
        if var2 is None:
            var2 = 0
        # --
    
        # Get names of ILP variables
        var1Name = var1
        var2Name = var2
        if not self.__varIsNumber(var1):
            var1Name = var1.VarName
        if not self.__varIsNumber(var2):
            var2Name = var2.VarName
            
        if onlyConstrains:
            m.addConstr(var1 >= var2)
            if self.ifLog: self.myLogger.debug("EQ created constraint only: %s => %s"%(var1Name, var2Name))
            
            m.addConstr(var1 <= var2)
            if self.ifLog: self.myLogger.debug("EQ created constraint only: %s <= %s"%(var1Name, var2Name))

            return
        
        varEQName = "epq_%s_%s"%(var1, var2)
        varEQName = varEQName[:254]
        varEQ = m.addVar(vtype=GRB.BINARY, name=varEQName)
            
        m.addConstr(var1 + var2 - varEQ <= 1)
        m.addConstr(var1 + var2 + varEQ >= 1)
        m.addConstr(-var1 + var2 + varEQ <= 1)
        m.addConstr(var1 - var2 + varEQ <= 1)
        
        m.update()
             
        if self.ifLog: self.myLogger.debug("EQ returns : %s"%(varEQ.VarName))
        return varEQ
    
    def countVar(self, m, *var, onlyConstrains = False, limitOp = 'None', limit = 1, logicMethodName = "COUNT"):
        BigM = 100
        
        if not limitOp:
            if self.ifLog: self.myLogger.error("%s called with no operation specified for comparing limit"%(logicMethodName))
            return None

        if limitOp not in ('<=', '>=', '=='):
            if self.ifLog: self.myLogger.error("%s called with incorrect operation specified for comparing limit %s"%(logicMethodName,limitOp))
            return None
            
        if self.ifLog: self.myLogger.debug("%s called with limit: %i and operation %s"%(logicMethodName,limit,limitOp))
        
        # -- Consider None
        varFixed = []  
        for v in var:
            if v is None:
                varFixed.append(0) # when None
            else:
                varFixed.append(v)
        
        var = varFixed
        # --
        
        varsInfo = self.preprocessLogicalMethodVar(varFixed, logicMethodName, logicMethodName,  minN=1)
            
        updatedLimit = limit - varsInfo['numberSum']
        
        # If only constructing constrains forcing OR to be true 
        if onlyConstrains:
            # Adding ILP constraint
            if limitOp == '>=': # ilp >= L
                if updatedLimit <= 0: # The constraint is satisfied - the limit is negative or zero
                    if self.ifLog: self.myLogger.debug("%s constraint is satisfied - the limit %i is negative or zero"%(logicMethodName,updatedLimit))
                    return
                elif updatedLimit > varsInfo['No_of_ilp']: # The limit is greater than the number of ILP variable - the constraint cannot be satisfied
                    raise Exception("ILP model is infeasible - %s limit %i is greater than the number of ILP variable %i - the constraint %s cannot be satisfied"
                                    %(logicMethodName,updatedLimit,varsInfo['No_of_ilp'],logicMethodName))
                else:
                    # Create Constraint
                    m.addConstr(varsInfo['varSumLinExpr'] >= updatedLimit, name='Count %s:'%(logicMethodName)) # varSumLinExpr >= updatedLimit
                    if self.ifLog: self.myLogger.debug("%s created ILP constraint: %s >= %i"%(logicMethodName,varsInfo['varSumLinExprStr'],updatedLimit))
                    
            # This check is common for '<=' and '=='
            elif updatedLimit < 0: # The constraint not is satisfied - the limit is negative or zero so ilp sum cannot be less than it - ilp sum is zero or more
                raise Exception("ILP model is infeasible - %s limit %i is negative or zero, ilp sum cannot be less than it - the constraint %s cannot be satisfied"
                                    %(logicMethodName,updatedLimit,logicMethodName))
                
            elif limitOp == '<=': # ilp <= L
                if varsInfo['No_of_ilp'] == 0: # sum Ilp =0 and L >= 0
                    if self.ifLog: self.myLogger.debug("%s constraint is satisfied - no ILP variable"%(logicMethodName))
                    return
                else:
                    m.addConstr(varsInfo['varSumLinExpr'] <= updatedLimit, name='Count %s:'%(logicMethodName)) # varSumLinExpr <= updatedLimit
                    if self.ifLog: self.myLogger.debug("%s created ILP constraint: %s <= %i"%(logicMethodName,varsInfo['varSumLinExprStr'],updatedLimit))

            elif limitOp == '==': # ilp == L
                if varsInfo['No_of_ilp'] == 0:
                    if updatedLimit == 0:
                        if self.ifLog: self.myLogger.debug("%s constraint is satisfied - no ILP variable"%(logicMethodName))
                        return
                    else: # updatedLimit > 0
                        raise Exception("ILP model is infeasible - %s limit %i is not zero as number of ILP variable is zero - the constraint %s cannot be satisfied"
                                    %(logicMethodName,updatedLimit,logicMethodName))
                else:  
                    m.addConstr(varsInfo['varSumLinExpr'] == updatedLimit, name='Count %s:'%(logicMethodName)) # varSumLinExpr == updatedLimit
                    if self.ifLog: self.myLogger.debug("%s created ILP constraint: %s == %i"%(logicMethodName,varsInfo['varSumLinExprStr'],updatedLimit))

            return
        
        # ------- If creating variables representing value of OR build of provided variables
        else:
        # Build constrains
            if limitOp == '>=': # ilp >= L
                if updatedLimit <= 0: # The constraint is satisfied - the limit is negative or zero - return 1
                    if self.ifLog: self.myLogger.debug("%s constraint is satisfied - the limit %i is negative or zero - return 1"%(logicMethodName,updatedLimit))
                    return 1
                elif updatedLimit > varsInfo['No_of_ilp']: # The limit is greater than the number of ILP variable - the constraint cannot be satisfied
                    if self.ifLog: self.myLogger.debug("%s limit %i is greater than the number of ILP variable %i - the constraint %s cannot be satisfied - return False"
                                    %(logicMethodName,updatedLimit,varsInfo['No_of_ilp'],logicMethodName))
                    return False
                else:
                    # Create new variable
                    varCOUNT = m.addVar(vtype=GRB.BINARY, name=varsInfo['varName'])
                    if m: m.update()
                    
                    m.addConstr(varsInfo['varSumLinExpr'] - BigM *varCOUNT >= updatedLimit - BigM, name='Count %s:'%(logicMethodName))
                    m.addConstr(varsInfo['varSumLinExpr'] - BigM *varCOUNT <= updatedLimit, name='Count %s:'%(logicMethodName))   
                                 
            # This check is common for '<=' and '=='
            elif updatedLimit < 0: # The constraint not is satisfied - the limit is negative or zero so ilp sum cannot be less than it - ilp sum is zero or more
                if self.ifLog: self.myLogger.debug("%s limit %i is negative or zero, ilp sum cannot be less than it - the constraint %s cannot be satisfied - return False"
                                    %(logicMethodName,updatedLimit,logicMethodName))
                return False
                
            elif limitOp == '<=': # ilp <= L
                if varsInfo['No_of_ilp'] == 0: # No ILP variable - sum Ilp =0 and L >= 0
                    if self.ifLog: self.myLogger.debug("%s constraint is satisfied - no ILP variable"%(logicMethodName))
                    return True
                else:  
                    # Create new variable
                    varCOUNT = m.addVar(vtype=GRB.BINARY, name=varsInfo['varName'])
                    if m: m.update()
                    
                    m.addConstr(varsInfo['varSumLinExpr'] + BigM *varCOUNT <= updatedLimit + BigM, name='Count %s:'%(logicMethodName))
                    m.addConstr(varsInfo['varSumLinExpr'] + BigM *varCOUNT >= updatedLimit, name='Count %s:'%(logicMethodName))
                    
            elif limitOp == '==': # ilp == L
                if varsInfo['No_of_ilp'] == 0:
                    if updatedLimit == 0:
                        if self.ifLog: self.myLogger.debug("%s constraint is satisfied - no ILP variable - return True"%(logicMethodName))
                        return True
                    else: # updatedLimit > 0
                        if self.ifLog: self.myLogger.debug("I%s limit %i is not zero as number of ILP variable is zero - the constraint %s cannot be satisfied - return False"
                                    %(logicMethodName,updatedLimit,logicMethodName))
                        return False
                else:
                    # Create new variable
                    varCOUNT = m.addVar(vtype=GRB.BINARY, name=varsInfo['varName'])
                    if m: m.update()
            
                    m.addConstr(varsInfo['varSumLinExpr'] - updatedLimit <= BigM * (1 - varCOUNT), name='Count %s:'%(logicMethodName))
                    m.addConstr(varsInfo['varSumLinExpr'] - updatedLimit >= BigM * (varCOUNT - 1), name='Count %s:'%(logicMethodName))
    
            if self.ifLog: self.myLogger.debug("%s returns new variable: %s"%(logicMethodName,varsInfo['varName']))
            return varCOUNT
    
    def fixedVar(self, m, var, onlyConstrains = False): 
        logicMethodName = "FIXED"
        
        # -- Consider None
        if var is None: # not create Fixed constraint for None
            return None
        # --
                
        varName = var
        if not self.__varIsNumber(var):
            varName = var.VarName
            
        if self.ifLog: self.myLogger.debug("%s called with : %s"%(logicMethodName,varName))

        # If only constructing constrains 
       
        if self.__varIsNumber(var):
            self.myLogger.warning("%s has set value: %s - do nothing"%(logicMethodName,varName))
            return 
        
        fixedTag = None
        if var.VTag.startswith("True"):
            fixedTag = True
        elif var.VTag.startswith("False"):
            fixedTag = False
        
        if fixedTag == None: # Label in datanode was -100 
            return 1
        
        if fixedTag:    
            m.addConstr(var == 1, name='Fixed:')
            if self.ifLog: self.myLogger.debug("%s created constraint: Fixed %s == %i"%(logicMethodName,varName,1))

        elif not fixedTag:    
            m.addConstr(var == 0, name='Fixed:')
            if self.ifLog: self.myLogger.debug("%s created constraint: Fixed %s == %i"%(logicMethodName,varName,0))

        else:
            return # error
        
        if onlyConstrains:
            return
        
        if self.ifLog: self.myLogger.debug("%s returns: %i"%(logicMethodName,1))
        return 1


import abc
from collections import OrderedDict
from itertools import chain, permutations
from domiknows.graph.logicalConstrain import nandL

import torch

if __package__ is None or __package__ == '':
    from base import BaseGraphTree
    from concept import Concept
else:
    from .base import BaseGraphTree
    from .concept import Concept


class Transformed():
    def __init__(self, relation, property, fn=None):
        self.relation = relation
        if isinstance(property, (str, Relation)):
            property = self.relation.src[property]
        self.property = property
        self.fn = fn

    def __call__(self, data_item, device=None):
        value = self.property(data_item)
        try:
            mapping = self.relation.dst[self.relation](data_item)
        except KeyError:
            mapping = self.relation.src[self.relation.reversed](data_item).T
        mapping = mapping.to(dtype=torch.float, device=device)
        value = value.to(dtype=torch.float, device=device)
        if self.fn is None:
            return mapping.matmul(value)
        # mapping (N,M)
        # value (M,...)
        mapping = mapping.view(*(mapping.shape + (1,)*(len(value.shape)-1)))  # (N,M,...)
        value = value.unsqueeze(dim=0)  # (1,M,...)
        return self.fn(mapping * value)


@BaseGraphTree.localize_namespace
class Relation(BaseGraphTree):
    @classmethod
    def name(cls):  # complicated to use class property, just function
        return cls.__name__

    def __init__(self, src, dst, argument_name, reverse_of=None, auto_constraint=None):
        cls = type(self)
        if isinstance(argument_name, str):
            name = argument_name
        else:
            name = '{}-{}-{}-{}'.format(src.name, cls.name(), argument_name, dst.name)
        BaseGraphTree.__init__(self, name)
        self.src = src
        self.dst = dst
        if reverse_of is None:
            self.is_reversed = False
            src._out.setdefault(cls.name(), []).append(self)
            dst._in.setdefault(cls.name(), []).append(self)
            reverse_of = Relation(dst, src, f'{name}.reversed', self)
        else:
            self.is_reversed = True
        self.reversed = reverse_of
        self.auto_constraint = auto_constraint

    @property
    def mode(self):
        return 'backward' if self.is_reversed else 'forward'

    @property
    def auto_constraint(self):
        if self._auto_constraint is None and self.sup is not None:
            return self.sup.auto_constraint
        return self._auto_constraint or False  # if None, return False instead

    @auto_constraint.setter
    def auto_constraint(self, value):
        self._auto_constraint = value

    def __call__(self, *props, fn=None):
        return Transformed(self, props[0], fn=fn)
        # TODO: support mapping multiple props together?
        # for prop in props:
        #     assert prop.sup == self.src
        #     yield Transformed(self.relation, prop, fn=fn)

    @property
    def src(self):
        return self._src

    @src.setter
    def src(self, src):
        if isinstance(src, Concept):
            self._src = src
        else:
            raise TypeError('Unsupported type of relation src {} in relation. Supported type is Concept.'
                            .format(type(src)))

    @property
    def dst(self):
        return self._dst

    @dst.setter
    def dst(self, dst):
        self._dst = dst

    def what(self):
        return {'src': self.src, 'dst': self.dst}

    __metaclass__ = abc.ABCMeta

    def set_apply(self, name, sub):
        from ..sensor import Sensor
        from .property import Property
        if isinstance(sub, Property):
            # call usually come from attach, further from constructor of property
            BaseGraphTree.set_apply(self, name, sub)
        elif isinstance(sub, Sensor):
            if name not in self:
                with self:
                    prop = Property(prop_name=name)
            self.get_apply(name).attach(sub)


class OTORelation(Relation):
    pass


class OTMRelation(Relation):
    pass


class MTORelation(Relation):
    pass


class MTMRelation(Relation):
    pass


@Concept.relation_type('is_a')
class IsA(OTORelation):
    pass


@Concept.relation_type('not_a')
class NotA(OTORelation):
    def __init__(self, src, dst, *args, **kwargs):
        super().__init__(src, dst, *args, **kwargs)
        nandL(src, dst)


def disjoint(*concepts):
    rels = []
    for c1, c2 in permutations(concepts, r=2):
        rels.extend(c1.not_a(c2))
    return rels


@Concept.relation_type('has_a')
class HasA(MTORelation):
    pass


@Concept.relation_type('has_many')
class HasMany(OTMRelation):
    pass

@Concept.relation_type('contains')
class Contains(OTMRelation):
    pass

@Concept.relation_type('equal')
class Equal(OTORelation):
    pass



from typing import List, Dict, Tuple, NoReturn, Any
import torch
from ...graph import Property
from .. import Sensor, Learner
from .module import WrapperModule


class AllenNlpSensor(Sensor):
    def __init__(
        self,
        *pres: List[Property],
        output_only: bool=False
    ) -> NoReturn:
        Sensor.__init__(self)
        self.pres = pres
        self.output_only = output_only
        self.pre_dims = []
        for pre in pres:
            for sensor in pre.find(AllenNlpSensor):
                dim = sensor.output_dim
                self.pre_dims.append(dim)
                break
            else:
                raise RuntimeError('Could not determin input dim for {} from pre-requirement {}'.format(self.fullname, pre.fullname))

    def update_context(
        self,
        data_item: Dict[str, Any],
        force=False
    ) -> Dict[str, Any]:
        # update prerequired first
        for pre in self.pres:
            if not isinstance(pre, dict):
                raise RuntimeError('{}'.format(self.fullname))
            for name, sensor in pre.items():
                # choose one result to update finally
                # TODO: consider priority or confidence or merge somehow
                if isinstance(sensor, AllenNlpSensor) and not sensor.output_only:
                    sensor(data_item)
                    break
            else:  # no break
                raise RuntimeError('Not able to find a sensor for {} as prereqiured by {}'.format(
                    pre.fullname, self.fullname))

        # then call forward
        Sensor.update_context(self, data_item, force)

    @property
    def output_dim(self):
        return None

class ReaderSensor(AllenNlpSensor):
    def __init__(
        self,
        reader,
        key: str,
        output_only: bool=False
    ) -> NoReturn:
        AllenNlpSensor.__init__(self, output_only=output_only) # *pres=[]
        self.key = key
        self.reader = reader
        reader.claim(key, self)

    def forward(
        self,
        data_item: Dict[str, Any]
    ) -> Any:
        return data_item[self.fullname]



class ModuleSensor(AllenNlpSensor):
    def __init__(
        self,
        *pres: List[Property],
        output_only: bool=False
    ) -> NoReturn:
        AllenNlpSensor.__init__(self, *pres, output_only=output_only)
        self._module = None

    def create_module(self):
        raise NotImplementedError('Implemented of create_module is required in subclass of ModuleSensor.')

    @property
    def module(self):
        if self._module is None:
            module = self.create_module()
            self._module = WrapperModule(module, self.pres)
        return self._module

    @property
    def output_dim(self):
        return (self.module.get_output_dim(),)


class AllenNlpLearner(ModuleSensor, Learner):
    def parameters(self):
        return self.module.parameters()


class PreArgsModuleSensor(ModuleSensor):
    def forward(
        self,
        data_item: Dict[str, Any]
    ) -> Any:
        try:
            return self.module(*(data_item[pre.fullname] for pre in self.pres))
        except:
            print('Error during forward with sensor {}'.format(self.fullname))
            print('module:', self.module)
            raise
        return None


class PreArgsModuleLearner(PreArgsModuleSensor, AllenNlpLearner):
    pass


class SinglePreSensor(AllenNlpSensor):
    def __init__(
        self,
        pre: Property,
        output_only: bool=False
    ) -> NoReturn:
        AllenNlpSensor.__init__(self, pre, output_only=output_only)

    @property
    def pre(self):
        return self.pres[0]

    @property
    def pre_dim(self):
        return self.pre_dims[0]


class SinglePreLearner(SinglePreSensor, PreArgsModuleLearner):
    def __init__(
        self,
        pre: Property,
        output_only: bool=False
    ) -> NoReturn:
        PreArgsModuleLearner.__init__(self, pre, output_only=output_only)


class MaskedSensor(AllenNlpSensor):
    def get_mask(self, data_item: Dict[str, Any]):
        raise NotImplementedError('Implement get_mask in subclass of MaskedSensor.')


class SinglePreMaskedSensor(SinglePreSensor, MaskedSensor):
    def get_mask(self, data_item: Dict[str, Any]):
        for sensor in self.pre.find(MaskedSensor):
            break
        else:
            print(self.pre)
            raise RuntimeError('{} require at least one pre-required sensor to be MaskedSensor.'.format(self.fullname))
        return sensor.get_mask(data_item)


class SinglePreMaskedLearner(SinglePreMaskedSensor, SinglePreLearner):
    def forward(
        self,
        data_item: Dict[str, Any]
    ) -> Any:
        return self.module(data_item[self.pre.fullname], self.get_mask(data_item))


class SinglePreArgMaskedPairSensor(PreArgsModuleSensor, SinglePreMaskedSensor):
    pass



if __package__ is None or __package__ == '':
    from domiknows.solver.ilpConfig import ilpConfig 
    from domiknows.solver.ilpOntSolver import ilpOntSolver
else:
    from .ilpConfig import ilpConfig 
    from .ilpOntSolver import ilpOntSolver

class ilpOntSolverFactory:
    __instances = {}
    __classes = {}
    
    @classmethod
    def getClass(cls, *SolverClasses):
        if SolverClasses not in cls.__classes:
            class ImplClass(*SolverClasses): pass
            cls.__classes[SolverClasses] = ImplClass
        return cls.__classes[SolverClasses]

    @classmethod
    def getOntSolverInstance(cls, graph, *SupplementalClasses, _ilpConfig=ilpConfig, **kwargs) -> ilpOntSolver:
        if graph is None:
            return None
        
        if not isinstance(graph, set):
            graph = {graph}
            
        ontologies = []
        reuse_model = False
        
        graphOntologyError = set()
        for currentGraph in graph:
            if (hasattr(currentGraph, 'ontology')) and (currentGraph.ontology is not None) and (currentGraph.ontology.iri is not None):
                if currentGraph.ontology not in ontologies:
                    ontologies.append(currentGraph.ontology)
            else:
                graphOntologyError.add(currentGraph)
                
            if currentGraph.reuse_model:
                reuse_model = True
                
        ontologiesTuple = (*ontologies, )
        
        if _ilpConfig is not None:
            if _ilpConfig['ilpSolver'] == "Gurobi":
                if reuse_model:
                    kwargs['reuse_model'] = True

                if __package__ is None or __package__ == '':
                    from domiknows.solver.gurobiILPOntSolver import gurobiILPOntSolver
                else:
                    from .gurobiILPOntSolver import gurobiILPOntSolver
                SolverClass = cls.getClass(gurobiILPOntSolver, *SupplementalClasses)
            elif _ilpConfig['ilpSolver'] == "Gurobi1":
                if reuse_model:
                    kwargs['reuse_model'] = True

                if __package__ is None or __package__ == '':
                    from domiknows.solver.gurobiILPOntSolver1 import gurobiILPOntSolver
                else:
                    from .gurobiILPOntSolver1 import gurobiILPOntSolver
                SolverClass = cls.getClass(gurobiILPOntSolver, *SupplementalClasses)
            elif _ilpConfig['ilpSolver'] == "GEKKO":
                if __package__ is None or __package__ == '':
                    from domiknows.solver.gekkoILPOntSolver import gekkoILPOntSolver
                else:
                    from .gekkoILPOntSolver import gekkoILPOntSolver
                SolverClass = cls.getClass(gekkoILPOntSolver, *SupplementalClasses)
            elif _ilpConfig['ilpSolver'] == "mini_prob_debug":
                if __package__ is None or __package__ == '':
                    from domiknows.solver.mini_solver_debug import MiniProbSolverDebug
                else:
                    from .mini_solver_debug import MiniProbSolverDebug
                SolverClass = cls.getClass(MiniProbSolverDebug, *SupplementalClasses)
            elif _ilpConfig['ilpSolver'] == "mini_debug":
                if __package__ is None or __package__ == '':
                    from domiknows.solver.mini_solver_debug import MiniSolverDebug
                else:
                    from .mini_solver_debug import MiniSolverDebug
                SolverClass = cls.getClass(MiniSolverDebug, *SupplementalClasses)
            else:
                if __package__ is None or __package__ == '':
                    from domiknows.solver.dummyILPOntSolver import dummyILPOntSolver
                else:
                    from .dummyILPOntSolver import dummyILPOntSolver
                SolverClass = cls.getClass(dummyILPOntSolver, *SupplementalClasses)

        key = (SolverClass, ontologiesTuple, frozenset(_ilpConfig), frozenset(kwargs))
        if key not in cls.__instances:
            instance = SolverClass(graph, ontologiesTuple, _ilpConfig, **kwargs)
            cls.__instances[key] = instance

            if ontologiesTuple and graphOntologyError:
                for currentGraph in graphOntologyError:
                    instance.myLogger.error("Problem - graph %s ontology is not correctly defined %s"%(currentGraph.name,currentGraph.ontology))

            if ontologiesTuple:
                instance.myLogger.info("Returning new ilpOntSolver for %s using %s"%(ontologiesTuple,_ilpConfig['ilpSolver']))
            else:
                instance.myLogger.debug("Returning generic (not based on ontology) ilpOntSolver using %s"%(_ilpConfig['ilpSolver']))

            return instance
        else:
            instance = cls.__instances[key]
            if ontologiesTuple:
                instance.myLogger.info("Returning existing ilpOntSolver for %s using %s"%(ontologiesTuple,_ilpConfig['ilpSolver']))
            else:
                instance.myLogger.debug("Returning existing generic ilpOntSolver using %s"%(_ilpConfig['ilpSolver']))

            return instance



from itertools import combinations

import torch

from ...graph import Property
from ...sensor import Sensor, Learner
from ...sensor.torch.sensor import TorchSensor
from ...sensor.torch.learner import ModuleLearner

from .base import Mode


def all_properties(node):
    if isinstance(node, Property):
        return node
    return None


class BaseModel(torch.nn.Module):
    def __init__(self, graph, loss=None, metric=None, device='auto'):
        super().__init__()
        self.graph = graph
        self.loss = loss
        self.metric = metric
        self.mode_ = Mode.TRAIN
        self.device = device

    def mode(self, mode):
        self.mode_ = mode

    def reset(self):
        if self.loss is not None:
            self.loss.reset()
        if self.metric is not None:
            self.metric.reset()

    def move(self, value, device=None):
        parameters = list(self.parameters())
        if parameters:
            device = device or next(self.parameters()).device
        else:
            device = device
        if isinstance(value, torch.Tensor):
            return value.to(device)
        elif isinstance(value, list):
            return [self.move(v, device) for v in value]
        elif isinstance(value, tuple):
            return (self.move(v, device) for v in value)
        elif isinstance(value, dict):
            return {k: self.move(v, device) for k, v in value.items()}
        else:
            return value

    def forward(self, data_item):
        data_item = self.move(data_item)
        return data_item


class TorchModel(BaseModel):
    BaseSensor = TorchSensor
    BaseLearner = ModuleLearner

    def __init__(self, graph, loss=None, metric=None, device='auto'):
        super().__init__(graph, loss=loss, metric=metric, device=device)
        self.poi = {prop: (output_sensor, target_sensor) for prop, output_sensor, target_sensor in self.find_poi()}
        self.graph.poi = self.poi
        for node in self.graph.traversal_apply(all_properties):
            for sensor in node.find(self.BaseLearner):
                self.add_module(sensor.fullname, sensor.module)

    def find_poi(self):
        for prop in self.graph.traversal_apply(all_properties):
            for sensor1, sensor2 in combinations(prop.find(self.BaseSensor), r=2):
                if sensor1.target:
                    target_sensor = sensor1
                    output_sensor = sensor2
                elif sensor2.target:
                    target_sensor = sensor2
                    output_sensor = sensor1
                else:
                    # TODO: should different learners get closer?
                    continue
                if output_sensor.target:
                    # two targets, skip
                    continue
                yield prop, output_sensor, target_sensor


class PoiModel(TorchModel):
    def poi_loss(self, data_item, prop, output_sensor, target_sensor):
        if not self.loss:
            return 0
        logit = output_sensor(data_item)
        mask = output_sensor.mask(data_item)
        labels = target_sensor(data_item)

        local_loss = self.loss[output_sensor, target_sensor](logit, labels, mask)
        return local_loss

    def poi_metric(self, data_item, prop, output_sensor, target_sensor):
        if not self.metric:
            return None
        mask = output_sensor.mask(data_item)
        labels = target_sensor(data_item)
        inference = prop(data_item)

        local_metric = self.metric[output_sensor, target_sensor](inference, labels, mask)
        return local_metric

    def forward(self, data_item, inference=True):
        data_item = super().forward(data_item)
        loss = 0
        metric = {}

        for prop, (output_sensor, target_sensor) in self.poi.items():
            # make sure the sensors are evaluated
            output = output_sensor(data_item)
            target = target_sensor(data_item)
            # calculated any loss or metric
            loss += self.poi_loss(data_item, prop, output_sensor, target_sensor)
            metric[output_sensor, target_sensor] = self.poi_metric(data_item, prop, output_sensor, target_sensor)

        return loss, metric, data_item



from .program import LearningBasedProgram
from .model.pytorch import PoiModel, PoiModelToWorkWithLearnerWithLoss, SolverModel, SolverModelDictLoss
from .model.iml import IMLModel


class POIProgram(LearningBasedProgram):
    def __init__(self, graph, **kwargs):
        super().__init__(graph, PoiModel, **kwargs)


class SolverPOIProgram(LearningBasedProgram):
    def __init__(self, graph, **kwargs):
        super().__init__(graph, SolverModel, **kwargs)
        
        
class SolverPOIDictLossProgram(LearningBasedProgram):
    def __init__(self, graph, **kwargs):
        super().__init__(graph, SolverModelDictLoss, **kwargs)


class IMLProgram(LearningBasedProgram):
    def __init__(self, graph, **kwargs):
        super().__init__(graph, IMLModel, **kwargs)


class POILossProgram(LearningBasedProgram):
    def __init__(self, graph, poi=None):
        super().__init__(graph, PoiModelToWorkWithLearnerWithLoss, poi=poi)



import abc
from typing import Any, Dict
from itertools import chain
from domiknows.sensor.sensor import Sensor
from domiknows.graph.property import Property
from domiknows.utils import consume

import torch


class SkipSensor(Exception):
    pass


class TorchSensor(Sensor):
    def __init__(self, target=False):
        super().__init__()
        self.target = target

    def propagate_context(self, data_item, node, force=False):
        if not self.target:  # avoid target to be mixed in forward calculation
            super().propagate_context(data_item, node, force)

    def mask(self, data_item: Dict[str, Any]) -> Any:
        # allow safely skip mask
        raise SkipSensor


class DataSensor(TorchSensor):
    def __init__(self, key, target=False):
        super().__init__(target=target)
        self.key = key

    def forward(self, data_item):
        return data_item[self.key]


class LabelSensor(DataSensor):
    def __init__(self, key):
        super().__init__(key, target=True)


class FunctionalSensor(TorchSensor):
    def __init__(self, *pres, target=False):
        super().__init__(target)
        self.pres = pres

    def get_args(self, data_item, skip_none_prop=False, sensor_fn=None, sensor_filter=None):
        for pre in self.pres:
            if isinstance(pre, Property):
                for _, sensor in pre.items():
                    if sensor_filter and not sensor_filter(sensor):
                        continue
                    try:
                        if sensor_fn:
                            yield sensor_fn(sensor, data_item)
                        else:
                            yield sensor(data_item)
                        break
                    except SkipSensor:
                        pass
                else:  # no break
                    raise RuntimeError('Not able to find a sensor for {} as prereqiured by {}'.format(
                        pre.fullname, self.fullname))
            else:
                if not skip_none_prop:
                    yield pre

    @abc.abstractmethod
    def forward_func(self, *args, **kwargs):
        raise NotImplementedError

    def forward(
        self,
        data_item: Dict[str, Any],
    ) -> Dict[str, Any]:
        args = self.get_args(data_item, sensor_filter=lambda s: not s.target)
        return self.forward_func(*args)

    def mask(self, data_item):
        masks = list(self.get_args(data_item, sensor_fn=lambda s, c: s.mask(c)))
        masks_num = len(masks)
        mask = masks[0]
        mask = mask.float()
        for i in range(1, masks_num):
            for j in range(i, masks_num):
                masks[j].unsqueeze_(-2)
            masks[i] = masks[i].float()
            mask = mask.unsqueeze_(-1).matmul(masks[i])
        return mask


class CartesianSensor(FunctionalSensor):
    def forward_func(self, *inputs):
        # torch cat is not broadcasting, do repeat manually
        input_iter = iter(inputs)
        output = next(input_iter)
        for input in input_iter:
            dob, *dol, dof = output.shape
            dib, *dil, dif = input.shape
            assert dob == dib
            output = output.view(dob, *dol, *(1,)*len(dil), dof).repeat(1, *(1,)*len(dol), *dil, 1)
            input = input.view(dib, *(1,)*len(dol), *dil, dif).repeat(1, *dol, *(1,)*len(dil), 1)
            output = torch.cat((output, input), dim=-1)
        return output


from collections import OrderedDict, Counter

class NorminalSensor(FunctionalSensor):
    UNK = '__UNK__'
    SPECIAL_TOKENS = [UNK]

    def __init__(self, *pres, vocab_counter=None, least_count=0, special_tokens=[], target=False):
        super().__init__(*pres, target=target)
        self.least_count = least_count
        self.special_tokens = self.SPECIAL_TOKENS.copy()
        self.special_tokens.extend(special_tokens)
        self.fixed = False
        self._vocab = None
        self.set_vocab(vocab_counter)

    def set_vocab(self, vocab_counter=None):
        self.vocab_counter = vocab_counter or Counter()
        self.update_vocab(fixed=True)

    def update_counter(self, outputs):
        comsum(self.vocab_counter.update, chain(*outputs))

    def update_vocab(self, fixed=None):
        if self.fixed:
            raise RuntimeError('Vocab of {} has been fixed!'.format(self))
        vocab_list = self.special_tokens.copy()
        vocab_count = list(filter(lambda kv: kv[1] > self.least_count, self.vocab_counter))
        if vocab_count:
            vocab_to_add, _ = zip(*vocab_count)
        else:
            vocab_to_add = []
        vocab_list.extend(vocab_to_add)
        self._vocab = OrderedDict((token, idx) for idx, token in enumerate(vocab_list))
        if fixed is not None:
            self.fixed = fixed

    def encode(self, output):
        return self._vocab.get(output, self._vocab[self.UNK])

    def forward(self, data_item):
        raw_outputs = super().forward(data_item)
        if not self.fixed:
            self.update_counter(raw_outputs)
        encoded_outputs = [list(map(self.encode, raw_output)) for raw_output in raw_outputs]
        max_len = max(map(len, encoded_outputs))
        def pad(output):
            return torch.cat((
                torch.tensor(output, dtype=torch.long), 
                torch.zeros(max_len - len(output), dtype=torch.long)
                ))
        encoded_tensor = torch.stack(tuple(map(pad, encoded_outputs)))
        return encoded_tensor, raw_outputs

class SpacyTokenizorSensor(NorminalSensor):
    from spacy.lang.en import English
    nlp = English()

    def forward_func(self, sentences):
        tokens = self.nlp.tokenizer.pipe(sentences)
        return list(tokens)


TRANSFORMER_MODEL = 'bert-base-uncased'

class BertTokenizorSensor(FunctionalSensor):
    from transformers import BertTokenizer
    tokenizer = BertTokenizer.from_pretrained(TRANSFORMER_MODEL)

    def forward_func(self, sentences):
        tokens = self.tokenizer.batch_encode_plus(
            sentences,
            return_tensors='pt',
            return_attention_mask=True,
            #return_offsets_mapping=True,
        )
        tokens['tokens'] = self.tokenizer.convert_ids_to_tokens(tokens['input_ids'], skip_special_tokens=True)
        return tokens

class BertEmbeddingSensor(FunctionalSensor):
    from transformers import BertModel
    model = BertModel.from_pretrained(TRANSFORMER_MODEL)

    # {
    #     tokens: list[List[str]], sting based tokens
    #     input_ids: list[List[int]],
    #     -token_type_ids: list[List[int]] if return_token_type_ids is True (default)
    #     attention_mask: list[List[int]] if return_attention_mask is True (default)
    #     -overflowing_tokens: list[List[int]] if a ``max_length`` is specified and return_overflowing_tokens is True
    #     -num_truncated_tokens: List[int] if a ``max_length`` is specified and return_overflowing_tokens is True
    #     -special_tokens_mask: list[List[int]] if ``add_special_tokens`` if set to ``True`` and return_special_tokens_mask is True
    # }
    def forward_func(self, tokens):
        outputs = self.model(tokens['input_ids'])
        return outputs



import os
import logging
import pickle
from glob import glob
from collections import defaultdict
from itertools import chain
from typing import Dict, List, Tuple, Union
import torch
import numpy as np
import pandas as pd
from allennlp.data.vocabulary import Vocabulary
from allennlp.data.iterators import BucketIterator
from allennlp.common.params import Params
from allennlp.training import Trainer
from allennlp.training.optimizers import Optimizer
from allennlp.training.learning_rate_schedulers import LearningRateScheduler
from allennlp.nn.util import device_mapping
from ..trial import Trial
from ...utils import WrapperMetaClass
from ...solver import ilpOntSolverFactory
from ...solver.ilpOntSolverFactory import ilpOntSolverFactory
from ...solver.ilpOntSolver import ilpOntSolver
from ...solver.allennlpInferenceSolver import AllennlpInferenceSolver
from ...solver.allennlplogInferenceSolver import AllennlplogInferenceSolver
from ...sensor.allennlp.base import ReaderSensor
from ...sensor.allennlp.learner import SentenceEmbedderLearner
from ...sensor.allennlp.sensor import SentenceEmbedderSensor
from .. import Graph, Property
from ..dataNode import DataNode
from .model import GraphModel
from .utils import evaluate


DEBUG_TRAINING = 'REGR_DEBUG' in os.environ and os.environ['REGR_DEBUG']


class AllenNlpGraph(Graph, metaclass=WrapperMetaClass):
    __metaclass__ = WrapperMetaClass

    def __init__(
        self,
        *args,
        log_solver=False,
        **kwargs
    ):
        vocab = None # Vocabulary()
        self.model = GraphModel(self, vocab, *args, **kwargs)
        if log_solver:
            self.solver = ilpOntSolverFactory.getOntSolverInstance(self, AllennlplogInferenceSolver)
        else:
            self.solver = ilpOntSolverFactory.getOntSolverInstance(self, AllennlpInferenceSolver)
        self.solver_log_to(None)
        # do not invoke super().__init__() here

    def get_multiassign(self):
        def func(node):
            if isinstance(node, Property) and len(node) > 1:
                return node
            return None
        return list(self.traversal_apply(func))

    @property
    def poi(self):
        return self.get_multiassign()

    def get_sensors(self, *tests):
        def func(node):
            if isinstance(node, Property):
                yield from node.find(*tests)
        return list(self.traversal_apply(func))

    def solver_log_to(self, log_path:str=None):
        #solver_logger = logging.getLogger(ilpOntSolver.__module__)
        solver_logger = self.solver.myLogger
        solver_logger.propagate = False
        if DEBUG_TRAINING:
            solver_logger.setLevel(logging.DEBUG)
            param = {'precision': 4,
                     'threshold': np.inf,
                     'edgeitems': 3,
                     'linewidth': np.inf
                    }
            torch.set_printoptions(**param)
            np.set_printoptions(**param)
            pd.options.display.max_rows = None
            pd.options.display.max_columns = None
        else:
            solver_logger.setLevel(logging.INFO)
        solver_logger.handlers = []
        if log_path is not None:
            if DEBUG_TRAINING:
                handler = logging.FileHandler(log_path)
            else:
                logFilesize = 5 * 1024 * 1024 * 1024 # 5G
                logBackupCount = 4
                logFileMode = 'a'
                handler = logging.handlers.RotatingFileHandler(log_path,
                                                               mode=logFileMode,
                                                               maxBytes=logFilesize,
                                                               backupCount=logBackupCount,
                                                               encoding=None,
                                                               delay=0)
            solver_logger.addHandler(handler)

    def save(self, path, **objects):
        if not os.path.exists(path):
            os.makedirs(path)
        with open(os.path.join(path, 'model.th'), 'wb') as fout:
            torch.save(self.model.state_dict(), fout)
        self.model.vocab.save_to_files(os.path.join(path, 'vocab'))
        for k, v in objects.items():
            with open(os.path.join(path, k + '.pkl'), 'wb') as fout:
                pickle.dump(v, fout)

    def load(self, path, model:Union[str, int]='default', vocab=None):
        model_file = None
        if isinstance(model, int):
            model_file = 'model_state_epoch_{}'.format(model)
        elif isinstance(model, str):
            if model.endswith('.th'):
                model_file = model
            elif model == 'best':
                model_file = 'best.th'
            elif model == 'last':
                model_file = sorted(glob(os.path.join(path, "model_state_epoch_*.th")), key=lambda p: os.path.getmtime(p)).pop()
            elif model == 'default':
                model_file = 'model.th'
        if model_file is None:
            raise ValueError(('`model` in `load()` must be one of the following: '
                              'an integer for epoch number, '
                              'a string name ends with ".th", '
                              'the string "best" for best model, '
                              'the string "last" for last saved model, '
                              'or the string "default" for just "model.th".'))
        vocab = vocab or os.path.join(path, 'vocab')

        if torch.cuda.is_available() and not DEBUG_TRAINING:
            device = 0
            self.model.cuda()
        else:
            device = -1

        #print('Loading vocab from {}'.format(vocab))
        self.model.vocab = Vocabulary.from_files(vocab)
        self.model.extend_embedder_vocab()
        #print('Loading model from {}'.format(model_file))
        with open(model_file, 'rb') as fin:
            self.model.load_state_dict(torch.load(fin, map_location=device_mapping(device)))

    @property
    def reader(self):
        readers = set(self.get_sensors(ReaderSensor))
        assert len(readers) == 1 # consider only 1 reader now
        return readers.pop()

    def train(self, data_config, train_config):
        reader = self.reader
        train_dataset = reader.read(os.path.join(data_config.relative_path, data_config.train_path), metas={'dataset_type':'train'})
        valid_dataset = reader.read(os.path.join(data_config.relative_path, data_config.valid_path), metas={'dataset_type':'valid'})
        self.update_vocab_from_instances(train_dataset + valid_dataset, train_config.pretrained_files)

        # prepare optimizer
        #print({n: p.size() for n, p in self.model.named_parameters()})
        optimizer = Optimizer.from_params(self.model.named_parameters(), Params(train_config.optimizer))

        # prepare scheduler
        if 'scheduler' in train_config:
            scheduler = LearningRateScheduler.from_params(optimizer, Params(train_config.scheduler))
        else:
            scheduler = None

        # prepare iterator
        sorting_keys = [(sensor.fullname, 'num_tokens') for sensor in self.get_sensors(SentenceEmbedderSensor)]
        iterator = BucketIterator(sorting_keys=sorting_keys,
                                  track_epoch=True,
                                  **train_config.iterator)
        iterator.index_with(self.model.vocab)

        # prepare model
        training_state = self.model.training
        self.model.train()

        trainer = self.get_trainer(train_dataset, valid_dataset,
                                   optimizer=optimizer,
                                   learning_rate_scheduler=scheduler,
                                   iterator=iterator,
                                   **train_config.trainer)

        if train_config.trainer.serialization_dir is not None:
            self.solver_log_to(os.path.join(train_config.trainer.serialization_dir, 'solver.log'))

        metrics = trainer.train()

        # restore model
        self.model.train(training_state)

        return metrics

    def get_trainer(
        self,
        train_dataset,
        valid_dataset,
        summary_interval=100,
        histogram_interval=100,
        should_log_parameter_statistics=True,
        should_log_learning_rate=True,
        **kwargs
    ) -> Trainer:
        # prepare GPU
        if torch.cuda.is_available() and not DEBUG_TRAINING:
            device = 0
            self.model.cuda()
        else:
            device = -1

        trainer = Trainer(model=self.model,
                          train_dataset=train_dataset,
                          validation_dataset=valid_dataset,
                          shuffle=not DEBUG_TRAINING,
                          cuda_device=device,
                          summary_interval=summary_interval,
                          histogram_interval=histogram_interval,
                          should_log_parameter_statistics=should_log_parameter_statistics,
                          should_log_learning_rate=should_log_learning_rate,
                          **kwargs)

        return trainer

    def update_vocab_from_instances(self, instances, pretrained_files=None):
        #import pdb; pdb.set_trace()
        vocab = Vocabulary.from_instances(instances, pretrained_files=pretrained_files)
        self.model.vocab = vocab
        self.model.extend_embedder_vocab()

    def test(self, data_path, log_path=None, batch=1, log_violation=False):
        # prepare GPU
        if torch.cuda.is_available() and not DEBUG_TRAINING:
            device = 0
            self.model.cuda()
        else:
            device = -1

        reader = self.reader
        dataset = reader.read(data_path, metas={'dataset_type':'test'})

        # prepare iterator
        sentence_sensors = self.get_sensors(SentenceEmbedderLearner)
        sorting_keys = [(sensor.fullname, 'num_tokens') for sensor in sentence_sensors]
        iterator = BucketIterator(batch_size=batch,
                                  sorting_keys=sorting_keys,
                                  track_epoch=True)
        iterator.index_with(self.model.vocab)

        # prepare model
        training_state = self.model.training
        self.model.eval()

        self.solver_log_to(log_path)
        metrics = evaluate(model=self.model,
                           instances=dataset,
                           data_iterator=iterator,
                           cuda_device=device,
                           batch_weight_key=None,
                           log_violation=log_violation)

        # restore model
        self.model.train(training_state)

        return metrics

    def populate(self, data_item, root, *concepts, query=None, label_fn=None, data_dict=None):
        trial = Trial()
        if data_dict is None:
            data_dict = defaultdict(list)

        def peel(node_value, repeat_unpeelable=False):
            # get a iterator for each instance in the fist level
            # {a:[1,2,3], b:[[4.1,4.2],[5.1,5.2],[6.1,6.2]], c:{d:[7,8,9]}}
            # ->
            # [{a:1, b:[4.1,4.2], c{d:7}}, {a:2, b:[5.1,5.2], c{d:8}}, {a:3, b:[6.1,6.2], c{d:9}}]
            if isinstance(node_value, (torch.Tensor, np.ndarray, List, Tuple)):
                return iter(node_value)
            elif isinstance(node_value, (Dict,)):
                pvs = []
                keys = []
                for k, v in node_value.items():
                    pv = peel(v, repeat_unpeelable)
                    if pv:
                        pvs.append(pv)
                        keys.append(k)
                return (dict(zip(keys, zpv)) for zpv in zip(*pvs))
            # TODO: fields like "loss" fron allennlp might not be possible to be divided
            # according to dim(/batch). Return None or return the identical one for all?
            elif repeat_unpeelable:
                return repeat(node_value)
            return None

        def contain(node, k):
            return k.startswith(node.fullname)

        def contained_by(node):
            return lambda item: contain(node, item[0])

        root_data = []
        sub_data_item = {}
        for k, v in filter(contained_by(self), data_item.items()):
            node = self[k[len(self.fullname)+1:]]
            if not isinstance(node, Property): # skip sensors and learners
                continue
            if contain(root, k):
                if not root_data:
                    for index, _ in enumerate(peel(v)):
                        dataNode = DataNode(instanceID=len(root_data),
                                            instanceValue=None,
                                            ontologyNode=root)
                        dataNode.index = index
                        root_data.append(dataNode)
                #import pdb; pdb.set_trace()
                for dataNode, pv in zip(root_data, peel(v)):
                    dataNode.__dict__[node.prop_name] = pv
            else:
                sub_data_item[k] = v
        if label_fn:
            for dataNode in root_data:
                prob = label_fn(dataNode)
                if prob is not None:
                    trial[root, dataNode] = prob
        data_dict[root].extend(root_data)
        for rel in root.contains():
            pass
            #for dataNode in data_dict[concept]:
            #    dataNode
#             for rel in concept.is_a():
#                 pass
#             for rel in concept.has_a():
#                 pass
        return trial, data_dict

#     @staticmethod
#     def peel(data_item, index):
#         # get a new data_item by selecting index for first dim
#         def peel_one(node_value):
#             # closure variable - index
#             if isinstance(node_value, (torch.Tensor, List, Tuple)):
#                 return node_value[index]
#             elif isinstance(node_value, (Dict,)):
#                 return {k, peel_one(v, index): for k, v: node_value.items()}
#             # TODO: fields like "loss" fron allennlp might not be possible to be divided
#             # according to dim(/batch). Return None or return the identical one for all?
#             return None
#         peeled = {}
#         for k, v in data_item:
#             v = peel_one(v)
#             if v is not None:
#                 peeled[k] = v
#         return peeled

#     @staticmethod
#     def peeled(data_item):
#         keys, values = zip(*data_item.items())
#         zip(values)
#         # get a new data_item by selecting index for first dim
#         def peel_one(node_value):
#             # closure variable - index
#             if isinstance(node_value, (torch.Tensor, List, Tuple)):
#                 return node_value[index]
#             elif isinstance(node_value, (Dict,)):
#                 return {k, peel_one(v, index): for k, v: node_value.items()}
#             # TODO: fields like "loss" fron allennlp might not be possible to be divided
#             # according to dim(/batch). Return None or return the identical one for all?
#             return None
#         peeled = {}
#         for k, v in data_item:
#             v = peel_one(v)
#             if v is not None:
#                 peeled[k] = v
#         return peeled

#     def populate(self, data_item, *concepts, batch=None, query=None, key='label'):
#         from ..concept import Concept
#         from ..property import Property
#         from ..dataNode import DataNode

#         trial = Trial()
#         data_dict = defaultdict()

#         def contained(item):
#             k = item[0]
#             return k.startswith(self.fullname)

#         def extract_batch_value(node_value):
#             if batch is None:
#                 return node_value
#             # closure variables - batch
#             if isinstance(node_value, (torch.Tensor, List, Tuple)):
#                 return node_value[batch]
#             elif isinstance(node_value, (Dict,)):
#                 return {k, extract_batch(v): for k, v: node_value.items()}
#             return None # fields like "loss" fron allennlp might not be divided according to batch, or return the identical one for all?

#         for node_key, node_value in filter(contained, data_item.items()):
#             node = self[node_key]
#             if isinstance(node, Property):
#                 concept_node = node.sup
#                 if concept_node not in data_dict: # TODO: limit to *concepts?
#                     # create the nodes
#                     # get rid of batch
#                     node_value = extract_batch_value(node_value)
#                     if node_value is None:
#                         # skip if get none for batch
#                         continue
#                     data_dict[concept_node] = []
#                     # TODO: estimate the number of argument
#                     # FIXME: suppose 1 here, that means a single loop
#                     for _ in node_value:
#                         dataNode = DataNode(instanceID=len(data_dict[concept_node]),
#                                             instanceValue=None,
#                                             ontologyNode=concept_node)
#                         data_dict[concept_node].append(dataNode)
#                 # add a property
#                 # FIXME: same problem as above `for _ in node_value`
#                 for dataNode, value in zip(data_dict[concept_node], node_value):
#                     dataNode.__dict__[node.name] = value
#                     if node.name == key:
#                         # TODO: how to inteprete the value here then?
#                         # logistic output have two dim where [1] is the logit for being True
#                         trial[concept_node, dataNode] = value[-1]

#         # wire up
#         for concept in concepts:
#             for rel in concept.contains():
#                 for dataNode in data_dict[concept]:
#                     dataNode
#             for rel in concept.is_a():
#                 pass
#             for rel in concept.has_a():
#                 pass

#         return trial, data_dict



import os
import logging
REGR_SOLVER = os.environ.get('REGR_SOLVER', 'Gurobi')

ilpConfig = {
    # variable controlling what ILP solver is used  - one of "Gurobi", "GEKKO", None
    'ilpSolver' : REGR_SOLVER,

    # Logging configuration for ilpOntSolver
    'ifLog': True,
    'log_name' : 'ilpOntSolver', 
    'log_level' : logging.INFO,
    'log_filename' : 'logs/ilpOntSolver',
    'log_filesize' : 5*1024*1024*1024,
    'log_backupCount' : 5,
    'log_fileMode' : 'a'
}


from enum import Enum


class SolverSession:
    class VTYPE(Enum):
        BIN = bool
        INT = int
        DEC = float

    class CTYPE(Enum):
        EQ = '=='
        LT = '<'
        LE = '<='
        GT = '>'
        GE = '>='

    class OTYPE(Enum):
        MAX = 'max'
        MIN = 'min'

    def var(self, vtype, lb, ub, name=None):
        raise NotImplementedError

    def constr(self, lhs, ctype, rhs, name=None):
        raise NotImplementedError

    def obj(self, otype, expr):
        raise NotImplementedError

    def optimize(self):
        raise NotImplementedError



from .base import BaseGraphShallowTree


@BaseGraphShallowTree.localize_namespace
class Property(BaseGraphShallowTree):
    def __init__(self, prop_name):
        cls = type(self)
        context = cls._context[-1]
        self.prop_name = prop_name
        #name = '{}-{}'.format(context.name, prop_name)
        BaseGraphShallowTree.__init__(self, prop_name)

    def attach_to_context(self, name=None):
        BaseGraphShallowTree.attach_to_context(self, self.prop_name)

    def get_fullname(self, delim='/'):
        if self.sup is None:
            return self.name  # note: when it is not nested in any context, use the auto name
        from .concept import Concept
        if isinstance(self.prop_name, Concept):
            prop_name = '<{}>'.format(self.prop_name.name)
        else:
            prop_name = self.prop_name
        return self.sup.get_fullname(delim) + delim + str(prop_name)

    def attach(self, sub):
        from ..sensor import Sensor
        if isinstance(sub, Sensor):
            BaseGraphShallowTree.attach(self, sub)
        else:
            raise TypeError(
                'Attach Sensor instance to Property, {} instance given.'.format(type(sub)))

    def find(self, *sensor_tests):
        sensor_tests = list(sensor_tests)
        for i in range(len(sensor_tests)):
            if isinstance(sensor_tests[i], type):
                sensor_type = sensor_tests[i] # need a new variable to avoid closure
                sensor_tests[i] = lambda s: isinstance(s, sensor_type)
            if not callable(sensor_tests[i]):
                raise ValueError('Unsupported feature.py {} to find.'.format(sensor_tests[i]))

        for name, sensor in self.items():
            for sensor_test in sensor_tests:
                if not sensor_test(sensor):
                    break
            else:
                from ..sensor import Sensor
                if isinstance(sensor, Sensor):
                    yield sensor

    def __call__(self, data_item):
        # make sure every sensor are visited
        for _, sensor in self.items():
            sensor(data_item)
        # and see what is lefted
        return data_item[self]



import logging
import torch
import numpy as np
from tqdm import tqdm

from .program import LearningBasedProgram, get_len
from ..utils import consume, entuple, detuple
from .model.lossModel import PrimalDualModel, SampleLosslModel
from .model.base import Mode

# Primal-dual need multiple backward through constraint loss.
# It requires retain_graph=True.
# Memory leak problem with `backward(create_graph=True)`
# https://github.com/pytorch/pytorch/issues/4661#issuecomment-596526199
# workaround with following functions based on `grad()`
def backward(loss, parameters):
    parameters = list(parameters)
    grads = torch.autograd.grad(outputs=loss, inputs=parameters, retain_graph=False)
    assert len(grads) == len(parameters)
    for grad, parameter in zip(grads, parameters):
        parameter.grad = grad

def unset_backward(parameters):
    for parameter in parameters:
        parameter.grad = None

def reverse_sign_grad(parameters, factor=-1.):
    for parameter in parameters:
        if parameter.grad is not None:
            parameter.grad = factor * parameter.grad

class LossProgram(LearningBasedProgram):
    DEFAULTCMODEL = PrimalDualModel

    logger = logging.getLogger(__name__)

    def __init__(self, graph, Model, CModel=None, beta=1, **kwargs):
        super().__init__(graph, Model, **kwargs)
        if CModel is None:
            CModel = self.DEFAULTCMODEL
            
        from inspect import signature
        cmodelSignature = signature(CModel.__init__)
        
        cmodelKwargs = {}
        for param in cmodelSignature.parameters.values():
            paramName = param.name
            if paramName in kwargs:
                cmodelKwargs[paramName] = kwargs[paramName]
        self.cmodel = CModel(graph, **cmodelKwargs)
        self.copt = None
        self.beta = beta

    def to(self, device):
        super().to(device=device)
        if self.device is not None:
            self.model.to(self.device)
            self.cmodel.device = self.device
            self.cmodel.to(self.device)

    def train(
        self,
        training_set,
        valid_set=None,
        test_set=None,
        # COptim=None,  # SGD only
        c_lr=0.05,
        c_momentum=0.9,
        c_warmup_iters=10,  # warmup
        c_freq=10,
        c_freq_increase=5,  # d
        c_freq_increase_freq=1,
        c_lr_decay=4,  # strategy
        c_lr_decay_param=1,  # param in the strategy
        **kwargs):
        
        # if COptim is None:
        #     COptim = Optim
        # if COptim is not None and list(self.model.parameters()):
        #     self.copt = COptim(self.model.parameters())
        if list(self.cmodel.parameters()):
            self.copt = torch.optim.Adam(self.cmodel.parameters(), lr=c_lr)
        else:
            self.copt = None
            
        # To provide a session cache for cross-epoch variables like iter-count
        c_session = {'iter':0, 'c_update_iter':0, 'c_update_freq':c_freq, 'c_update':0}  
        
        return super().train(
            training_set,
            valid_set=valid_set,
            test_set=test_set,
            c_lr=c_lr,
            c_warmup_iters=c_warmup_iters,
            c_freq_increase=c_freq_increase,
            c_freq_increase_freq=c_freq_increase_freq,
            c_lr_decay=c_lr_decay,
            c_lr_decay_param=c_lr_decay_param,
            c_session=c_session,
            **kwargs)
    
    def call_epoch(self, name, dataset, epoch_fn, **kwargs):
        if dataset is not None:
            self.logger.info(f'{name}:')
            desc = name if self.epoch is None else f'Epoch {self.epoch} {name}'

            consume(tqdm(epoch_fn(dataset, **kwargs), total=get_len(dataset), desc=desc))

            if self.model.loss:
                self.logger.info(' - loss:')
                self.logger.info(self.model.loss)

                metricName = 'loss'
                metricResult = self.model.loss
                    
                if self.dbUpdate is not None:
                    self.dbUpdate(desc, metricName, metricResult)
                    
            if self.cmodel.loss and self.cmodel.loss is not None:
                desc = name if self.epoch is None else f'Epoch {self.epoch} {name}'
                self.logger.info(' - Constraint loss:')
                self.logger.info(self.cmodel.loss)

                metricName = 'Constraint_loss'
                metricResult = self.cmodel.loss

                if self.dbUpdate is not None:
                    self.dbUpdate(desc, metricName, metricResult)

            ilpMetric = None
            softmaxMetric = None

            if self.model.metric:
                self.logger.info(' - metric:')
                for key, metric in self.model.metric.items():
                    self.logger.info(f' - - {key}')
                    self.logger.info(metric)
                    try:
                        self.f.write(f' - - {name}')
                        self.f.write(f' - - {key}')
                        self.f.write("\n")
                        self.f.write(str(metric))
                        self.f.write("\n")
                    except:
                        pass
                    metricName = key
                    metricResult = metric
                    if self.dbUpdate is not None:
                        self.dbUpdate(desc, metricName, metricResult)

                    if key == 'ILP':
                        ilpMetric = metric

                    if key == 'softmax':
                        softmaxMetric = metric
#         super().call_epoch(name=name, dataset=dataset, epoch_fn=epoch_fn, **kwargs)
        

    

    def train_epoch(
        self, dataset,
        c_lr=1,
        c_warmup_iters=10,  # warmup
        c_freq_increase=1,  # d
        c_freq_increase_freq=1,
        c_lr_decay=0,  # strategy
        c_lr_decay_param=1,
        c_session={},
        **kwargs):
        assert c_session
        from torch import autograd
        torch.autograd.set_detect_anomaly(False)
        self.model.mode(Mode.TRAIN)
#         self.cmodel.mode(Mode.TRAIN)
        iter = c_session['iter']
        c_update_iter = c_session['c_update_iter']
        c_update_freq = c_session['c_update_freq']
        c_update = c_session['c_update']
        self.model.train()
        self.model.reset()
        self.cmodel.train()
        self.cmodel.reset()
        for data in dataset:
            if self.opt is not None:
                self.opt.zero_grad()
            if self.copt is not None:
                self.copt.zero_grad()
            mloss, metric, *output = self.model(data)  # output = (datanode, builder)
            if iter < c_warmup_iters:
                loss = mloss
            else:
                closs, *_ = self.cmodel(output[1])
                loss = mloss + self.beta * closs
            if self.opt is not None and loss:
                loss.backward()
                self.opt.step()
                iter += 1
            if (
                self.copt is not None and
                loss and
                iter > c_warmup_iters and
                iter - c_update_iter > c_update_freq
            ):
                # NOTE: Based on the face the gradient of lambda in dual
                #       is reversing signs of their gradient in primal,
                #       we avoid a repeated backward (and also pytorch's
                #       retain_graph problem), we simply reverse the sign
                #       for the dual. Don't zero_grad() here!
                reverse_sign_grad(self.model.parameters())
                self.copt.step()
                # update counting
                c_update_iter = iter
                c_update += 1
                # update freq
                if c_freq_increase_freq > 0 and c_update % c_freq_increase_freq == 0:
                    c_update_freq += c_freq_increase
                # update c_lr
                if c_lr_decay == 0:
                    # on the paper
                    # c_lr_decay_param = beta
                    def update_lr(lr):
                        return c_lr * 1. / (1 + c_lr_decay_param * c_update)
                elif c_lr_decay == 1:
                    # in pd code / srl strategy 1
                    # c_lr_decay_param = lr_decay_after
                    def update_lr(lr):
                        return lr * np.sqrt(((c_update-1.) / c_lr_decay_param + 1.) / (c_update / c_lr_decay_param + 1.))
                elif c_lr_decay == 2:
                    # in pd code / srl strategy 2
                    # c_lr_decay_param = lr_decay_after
                    def update_lr(lr):
                        return lr * (((c_update-1.) / c_lr_decay_param + 1.) / (c_update / c_lr_decay_param + 1.))
                elif c_lr_decay == 3:
                    # in pd code / srl strategy 3
                    # c_lr_decay_param = lr_decay_after
                    assert c_lr_decay_param <= 1.
                    def update_lr(lr):
                        return lr * c_lr_decay_param
                elif c_lr_decay == 4:
                    # in pd code / ner strategy 1
                    def update_lr(lr):
                        return lr * np.sqrt((c_update+1) / (c_update+2))
                else:
                    raise ValueError(f'c_lr_decay={c_lr_decay} not supported.')
                for param_group in self.copt.param_groups:
                    param_group['lr'] = update_lr(param_group['lr'])
            yield (loss, metric, *output[:1])

        c_session['iter'] = iter
        c_session['c_update_iter'] = c_update_iter
        c_session['c_update_freq'] = c_update_freq
        c_session['c_update'] = c_update

    def test_epoch(self, dataset, **kwargs):
        # just to consum kwargs
#         self.cmodel.mode(Mode.TEST)
        yield from super().test_epoch(dataset)
        
    def populate_epoch(self, dataset):
        self.model.mode(Mode.POPULATE)
#         self.cmodel.mode(Mode.POPULATE)
        self.model.reset()
        with torch.no_grad():
            for i, data_item in enumerate(dataset):
                _, _, *output = self.model(data_item)
                yield detuple(*output[:1])
        
class PrimalDualProgram(LossProgram):
    logger = logging.getLogger(__name__)

    def __init__(self, graph, Model, beta=1, **kwargs):
        super().__init__(graph, Model, CModel=PrimalDualModel, beta=beta, **kwargs)
        
class SampleLossProgram(LossProgram):
    logger = logging.getLogger(__name__)

    def __init__(self, graph, Model, beta=1, **kwargs):
        super().__init__(graph, Model, CModel=SampleLosslModel, beta=beta, **kwargs)


    def train(
        self,
        training_set,
        valid_set=None,
        test_set=None,
        # COptim=None,  # SGD only
        c_lr=0.05,
        c_momentum=0.9,
        c_warmup_iters=10,  # warmup
        c_freq=10,
        c_freq_increase=5,  # d
        c_freq_increase_freq=1,
        c_lr_decay=4,  # strategy
        c_lr_decay_param=1,  # param in the strategy
        **kwargs):
        
        return super().train(
            training_set=training_set,
            valid_set=valid_set,
            test_set=test_set,
            c_lr=c_lr,
            c_momentum=c_momentum,
            c_warmup_iters=c_warmup_iters,  # warmup
            c_freq=c_freq,
            c_freq_increase=c_freq_increase,  # d
            c_freq_increase_freq=c_freq_increase_freq,
            c_lr_decay=c_lr_decay,  # strategy
            c_lr_decay_param=c_lr_decay_param,  # param in the strategy
            **kwargs)


    def train_epoch(
        self, dataset,
        c_warmup_iters=0,  # warmup
        c_session={},
        **kwargs):
        self.model.mode(Mode.TRAIN)
#         self.cmodel.mode(Mode.TRAIN)
        assert c_session
        iter = c_session['iter']
        self.model.train()
        self.model.reset()
        self.cmodel.train()
        self.cmodel.reset()
        for data in dataset:
            if self.opt is not None:
                self.opt.zero_grad()
            if self.copt is not None:
                self.copt.zero_grad()
            mloss, metric, *output = self.model(data)  # output = (datanode, builder)
            if iter < c_warmup_iters:
                loss = mloss
            else:
                closs, *_ = self.cmodel(output[1])
                if torch.is_tensor(closs):
                    loss = mloss + self.beta * closs
                else:
                    loss = mloss
                    
                if loss != loss:
                    raise Exception("Calculated loss is nan")
                
            if self.opt is not None and loss:
                loss.backward()
                # for name, param in self.model.named_parameters():
                #     if param.requires_grad:
                #         print (name, param.grad)

                self.opt.step()
                iter += 1
            
            if (
                self.copt is not None and
                loss
            ):
                self.copt.step()
            
            yield (loss, metric, *output[:1])

        c_session['iter'] = iter


import numpy as np

from domiknows.graph import Concept

from .ilpOntSolver import ilpOntSolver
from .session.solver_session import SolverSession
from .session.gurobi_session import GurobiSession
from .constructor.constructor import ScoreConstructor, ProbConstructor


class MiniSolverDebug(ilpOntSolver):
    ilpSolver = 'mini_debug'

    def __init__(self, graph, ontologiesTuple, _ilpConfig, constructor=None, SessionType=None, **kwargs):
        super().__init__(graph, ontologiesTuple, _ilpConfig)
        def func(node):
            if isinstance(node, Concept):
                return node
        self.names = {}
        for cur_graph in graph:
            self.names.update({concept.name: concept for concept in cur_graph.traversal_apply(func)})
        self.constructor = constructor or ScoreConstructor(**kwargs)
        self.SessionType = SessionType or GurobiSession

    def set_predication(self, predicate, idx, value):
        # predicates_result[concept][idx] = session.get_value(variables[concept, x])
        predicate[idx] = value

    def calculateILPSelection(self, data, *predicates_list):
        concepts_list = []
        for predicates in predicates_list:
            concept_dict = {}
            for predicate, v in predicates.items():
                concept = self.names[predicate]
                concept_dict[concept] = v
            concepts_list.append(concept_dict)

        # call to solve
        concepts_results = self.solve_legacy(data, *concepts_list)

        # prepare result
        results = [{k.name:v for k, v in concepts_result.items()}
                for concepts_result in concepts_results]
        return results

    def solve_legacy(self, data, *predicates_list):
        # data is a list of objects of the base type
        # predicates_list is a list of predicates
        # predicates_list[i] is the dict of predicates of the objects of the base type to the power of i
        # predicates_list[i][concept] is the prediction result of the predicate for concept
        self.myLogger.debug('Start for data %s', data)
        candidates = self.constructor.candidates(data, *predicates_list)

        session = self.SessionType()
        variables, predictions, variables_not, predictions_not = self.constructor.variables(session, candidates, * predicates_list)
        self.constructor.constraints(session, candidates, variables, variables_not, *predicates_list)

        # Set objective
        objective = self.constructor.objective(candidates, variables, predictions, variables_not, predictions_not, *predicates_list)
        session.obj(SolverSession.OTYPE.MAX, objective)
        self.myLogger.debug(' - model %s', session)

        # solve
        self.myLogger.debug('start optimize')
        session.optimize()
        self.myLogger.debug('end optimize')
        self.myLogger.debug(' - model %s', session)

        # collect result
        length = len(data)
        retval = []
        for arity, predicates in enumerate(predicates_list, 1):
            predicates_result = {}
            retval.append(predicates_result)
            for concept, _ in predicates.items():
                self.myLogger.debug('for %s', concept.name)
                predicates_result[concept] = np.zeros((length,) * arity)
                for x in candidates[concept]:
                    #import pdb; pdb.set_trace()
                    if (concept, x) not in variables: continue
                    # NB: candidates generated by 'C' order
                    idx, _ = zip(*x)
                    value = session.get_value(variables[concept, x])
                    self.set_predication(predicates_result[concept], idx, value)
                #import pdb;pdb.set_trace()
                self.myLogger.debug(predicates_result[concept])

        if len(retval) == 1:
            return retval[0]

        return retval

class MiniProbSolverDebug(MiniSolverDebug):
    ilpSolver = 'mini_prob_debug'

    def __init__(self, graph, ontologiesTuple, _ilpConfig, SessionType=None, **kwargs):
        constructor = ProbConstructor(**kwargs)
        super().__init__(graph, ontologiesTuple, _ilpConfig, constructor=constructor, SessionType=SessionType)



import json
import re


class RegrReader:
    def __init__(self, file, type="json"):
        self.file = file
        if type == "json":
            with open(file, 'r') as myfile:
                data = myfile.read()
            # parse file
            self.objects = json.loads(data)
        else:
            self.objects = self.parse_file()

    # you should return the objects list here
    def parse_file(self):
        pass

    def make_object(self, item):
        result = {}
        pattern = re.compile("^get.+val$")
        _list = [method_name for method_name in dir(self)
                 if callable(getattr(self, method_name)) and pattern.match(method_name)]
        for func in _list:
            name = func.replace("get", "", 1)
            k = name.rfind("val")
            name = name[:k]
            result[name] = getattr(self, func)(item)
        return result

    def run(self):
        for item in self.objects:
            yield self.make_object(item)

    def __iter__(self):
        for item in self.objects:
            yield self.make_object(item)

    def __len__(self):
        return len(self.objects)


from typing import Optional

from overrides import overrides
import torch
from sklearn import metrics

from allennlp.common.checks import ConfigurationError
from allennlp.training.metrics import Metric, F1Measure


@Metric.register("auc_fixed")
class Auc(Metric):
    """
    The AUC Metric measures the area under the receiver-operating characteristic
    (ROC) curve for binary classification problems.
    """

    def __init__(self, positive_label=1):
        super(Auc, self).__init__()
        self._positive_label = positive_label
        self._all_predictions = torch.FloatTensor()
        self._all_gold_labels = torch.LongTensor()

    def __call__(self,
                 predictions: torch.Tensor,
                 gold_labels: torch.Tensor,
                 mask: Optional[torch.Tensor] = None):
        """
        Parameters
        ----------
        predictions : ``torch.Tensor``, required.
            A one-dimensional tensor of prediction scores of shape (batch_size).
        gold_labels : ``torch.Tensor``, required.
            A one-dimensional label tensor of shape (batch_size), with {1, 0}
            entries for positive and negative class. If it's not binary,
            `positive_label` should be passed in the initialization.
        mask: ``torch.Tensor``, optional (default = None).
            A one-dimensional label tensor of shape (batch_size).
        """

        predictions, gold_labels = self.unwrap_to_tensors(
            predictions, gold_labels)

        # Sanity checks.
        if gold_labels.dim() != 1:
            raise ConfigurationError("gold_labels must be one-dimensional, "
                                     "but found tensor of shape: {}".format(gold_labels.size()))
        if predictions.dim() != 1:
            raise ConfigurationError("predictions must be one-dimensional, "
                                     "but found tensor of shape: {}".format(predictions.size()))

        unique_gold_labels = torch.unique(gold_labels)
        if unique_gold_labels.numel() > 2:
            raise ConfigurationError("AUC can be used for binary tasks only. gold_labels has {} unique labels, "
                                     "expected at maximum 2.".format(unique_gold_labels.numel()))

        gold_labels_is_binary = list(torch.sort(
            unique_gold_labels)[0].numpy()) <= [0, 1]
        if not gold_labels_is_binary and self._positive_label not in unique_gold_labels:
            raise ConfigurationError("gold_labels should be binary with 0 and 1 or initialized positive_label "
                                     "{} should be present in gold_labels".format(self._positive_label))

        if mask is None:
            batch_size = gold_labels.shape[0]
            mask = torch.ones(batch_size)
        mask = mask.byte()

        self._all_predictions = torch.cat([self._all_predictions,
                                           torch.masked_select(predictions, mask).float()], dim=0)
        self._all_gold_labels = torch.cat([self._all_gold_labels,
                                           torch.masked_select(gold_labels, mask).long()], dim=0)

    def get_metric(self, reset: bool = False):
        if self._all_gold_labels.shape[0] == 0:
            return 0.5
        if not any(self._all_gold_labels == self._positive_label):
            return float('nan')
        false_positive_rates, true_positive_rates, _ = metrics.roc_curve(self._all_gold_labels.numpy(),
                                                                         self._all_predictions.numpy(),
                                                                         pos_label=self._positive_label)
        auc = metrics.auc(false_positive_rates, true_positive_rates)
        if reset:
            self.reset()
        return auc

    @overrides
    def reset(self):
        self._all_predictions = torch.FloatTensor()
        self._all_gold_labels = torch.LongTensor()


@Metric.register("ap")
class AP(Metric):
    """
    The Average Presision measurement for binary classification problems.
    """

    def __init__(self):
        super(AP, self).__init__()
        self._all_predictions = torch.FloatTensor()
        self._all_gold_labels = torch.LongTensor()

    def __call__(self,
                 predictions: torch.Tensor,
                 gold_labels: torch.Tensor,
                 mask: Optional[torch.Tensor] = None):
        """
        Parameters
        ----------
        predictions : ``torch.Tensor``, required.
            A one-dimensional tensor of prediction scores of shape (batch_size).
        gold_labels : ``torch.Tensor``, required.
            A one-dimensional label tensor of shape (batch_size), with {1, 0}
            entries for positive and negative class. If it's not binary,
            `positive_label` should be passed in the initialization.
        mask: ``torch.Tensor``, optional (default = None).
            A one-dimensional label tensor of shape (batch_size).
        """

        predictions, gold_labels = self.unwrap_to_tensors(
            predictions, gold_labels)

        # Sanity checks.
        if gold_labels.dim() != 1:
            raise ConfigurationError("gold_labels must be one-dimensional, "
                                     "but found tensor of shape: {}".format(gold_labels.size()))
        if predictions.dim() != 1:
            raise ConfigurationError("predictions must be one-dimensional, "
                                     "but found tensor of shape: {}".format(predictions.size()))

        unique_gold_labels = torch.unique(gold_labels)
        if unique_gold_labels.numel() > 2:
            raise ConfigurationError("AUC can be used for binary tasks only. gold_labels has {} unique labels, "
                                     "expected at maximum 2.".format(unique_gold_labels.numel()))

        gold_labels_is_binary = list(torch.sort(
            unique_gold_labels)[0].numpy()) <= [0, 1]
        if not gold_labels_is_binary and self._positive_label not in unique_gold_labels:
            raise ConfigurationError("gold_labels should be binary with 0 and 1 or initialized positive_label "
                                     "{} should be present in gold_labels".format(self._positive_label))

        if mask is None:
            batch_size = gold_labels.shape[0]
            mask = torch.ones(batch_size)
        mask = mask.byte()

        self._all_predictions = torch.cat([self._all_predictions,
                                           torch.masked_select(predictions, mask).float()], dim=0)
        self._all_gold_labels = torch.cat([self._all_gold_labels,
                                           torch.masked_select(gold_labels, mask).long()], dim=0)

    def get_metric(self, reset: bool = False):
        if self._all_gold_labels.shape[0] == 0:
            return 0.5
        if sum(self._all_gold_labels) == 0:
            return float('nan')

        ap = metrics.average_precision_score(self._all_gold_labels.numpy(),
                                             self._all_predictions.numpy())
        if reset:
            self.reset()
        return ap

    @overrides
    def reset(self):
        self._all_predictions = torch.FloatTensor()
        self._all_gold_labels = torch.LongTensor()


@Metric.register("pr_auc")
class PRAuc(Metric):
    """
    The Average Presision measurement for binary classification problems.
    """

    def __init__(self):
        super(PRAuc, self).__init__()
        self._all_predictions = torch.FloatTensor()
        self._all_gold_labels = torch.LongTensor()

    def __call__(self,
                 predictions: torch.Tensor,
                 gold_labels: torch.Tensor,
                 mask: Optional[torch.Tensor] = None):
        """
        Parameters
        ----------
        predictions : ``torch.Tensor``, required.
            A one-dimensional tensor of prediction scores of shape (batch_size).
        gold_labels : ``torch.Tensor``, required.
            A one-dimensional label tensor of shape (batch_size), with {1, 0}
            entries for positive and negative class. If it's not binary,
            `positive_label` should be passed in the initialization.
        mask: ``torch.Tensor``, optional (default = None).
            A one-dimensional label tensor of shape (batch_size).
        """

        predictions, gold_labels = self.unwrap_to_tensors(
            predictions, gold_labels)

        # Sanity checks.
        if gold_labels.dim() != 1:
            raise ConfigurationError("gold_labels must be one-dimensional, "
                                     "but found tensor of shape: {}".format(gold_labels.size()))
        if predictions.dim() != 1:
            raise ConfigurationError("predictions must be one-dimensional, "
                                     "but found tensor of shape: {}".format(predictions.size()))

        unique_gold_labels = torch.unique(gold_labels)
        if unique_gold_labels.numel() > 2:
            raise ConfigurationError("AUC can be used for binary tasks only. gold_labels has {} unique labels, "
                                     "expected at maximum 2.".format(unique_gold_labels.numel()))

        gold_labels_is_binary = list(torch.sort(
            unique_gold_labels)[0].numpy()) <= [0, 1]
        if not gold_labels_is_binary and self._positive_label not in unique_gold_labels:
            raise ConfigurationError("gold_labels should be binary with 0 and 1 or initialized positive_label "
                                     "{} should be present in gold_labels".format(self._positive_label))

        if mask is None:
            batch_size = gold_labels.shape[0]
            mask = torch.ones(batch_size)
        mask = mask.byte()

        self._all_predictions = torch.cat([self._all_predictions,
                                           torch.masked_select(predictions, mask).float()], dim=0)
        self._all_gold_labels = torch.cat([self._all_gold_labels,
                                           torch.masked_select(gold_labels, mask).long()], dim=0)

    def get_metric(self, reset: bool = False):
        if sum(self._all_gold_labels) == 0:
            return float('nan')
        p, r, t = metrics.precision_recall_curve(self._all_gold_labels.numpy(),
                                                 self._all_predictions.numpy())
        auc = metrics.auc(r, p)

        if reset:
            self.reset()
        return auc

    @overrides
    def reset(self):
        self._all_predictions = torch.FloatTensor()
        self._all_gold_labels = torch.LongTensor()
from typing import Tuple


@Metric.register("precision")
class Precision(F1Measure):
    """
    Extract Precision from F1Measure.
    """

    def get_metric(self, reset: bool = False) -> float:
        metric = super().get_metric(reset=reset)
        return metric[0]

@Metric.register("epoch")
class Epoch(Metric):
    def __init__(self):
        super(Epoch, self).__init__()
        self.epoch_num = None

    def __call__(self, data):
        self.epoch_num = data['epoch_num'][0]  # (batch,) there are copies

    def get_metric(self, reset: bool = False):
        return self.epoch_num

@Metric.register("dataset_type")
class DatasetType(Metric):
    def __init__(self):
        super(DatasetType, self).__init__()
        self.dataset_type = None
        self.dataset_types = {}

    def __call__(self, data):
        self.dataset_type = data['dataset_type'][0]  # (batch,) there are copies
        if self.dataset_type not in self.dataset_types:
            self.dataset_types[self.dataset_type] = len(self.dataset_types)

    def get_metric(self, reset: bool = False):
        return self.dataset_types[self.dataset_type]



import abc

if __package__ is None or __package__ == '':
    from domiknows.solver.ilpConfig import ilpConfig 
else:
    from .ilpConfig import ilpConfig 
    
class ilpBooleanProcessor(object):
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, _ildConfig = ilpConfig) -> None:
        super().__init__()

    # NOT Negation
    # Create new variable varNOT, create constrain 1 - var == varNOT, return varNOT
    # 1 - var == varNOT
    # if onlyConstrains then only construct constrain 1 - var >= 0
    @abc.abstractmethod
    def notVar(self, m, _var, onlyConstrains = False): pass
    
    # AND Conjunction with 2 variable
    # Create new variable varAND, create constrains:
    # varAND <= var1
    # varAND <= var2 
    # var1 + var2 <= varAND + 2 - 1
    # return varAND
    # if onlyConstrains then only construct constrain var1 + var2 >= 2
    @abc.abstractmethod
    def and2Var(self, m, _var1, _var2,  onlyConstrains = False): pass
        
    # AND Conjunction
    # Create new variable varAND, create constrains:
    # varAND <= var1
    # varAND <= var2
    # ....
    # varAND <= varN
    # var1 + var2 + .. + varN <= varAND + N - 1
    # return varAND
    # if onlyConstrains then only construct constrain var1 + var2 + .. + varN >= N
    @abc.abstractmethod
    def andVar(self, m, *_var, onlyConstrains = False): pass
       
    # OR Disjunction with 2 variables
    # Create new variable varOR, create constrains:
    # var1 <= varOR
    # var2 <= varOR 
    # var1 + var2 >= varOR 
    # return varOR
    # if onlyConstrains then only construct constrain var1 + var2 >= 1
    #
    # if limit > 1
    # var1 + var2 - (limit - 1) >= varOR 
    # if onlyConstrains then only construct constrain var1 + var2 >= limit
    @abc.abstractmethod
    def or2Var(self, m, _var1, _var2,  onlyConstrains = False): pass
        
    # OR Disjunction
    # Create new variable varOR, create constrains:
    # var1 <= varOR
    # var2 <= varOR 
    # ...
    # varN <= varOR
    # var1 + var2 + ... + varN >= varOR
    # return varOR
    # if onlyConstrains then only construct constrain var1 + var2 + ... + varN >= 1
    # if limit > 1
    # var1 + var2 + ... + varN - (limit - 1) >= varOR 
    # if onlyConstrains then only construct constrain var1 + var2 + ... + varN >= limit
    @abc.abstractmethod
    def orVar(self, m, *_var, onlyConstrains = False): pass

    # NAND (Alternative denial) with 2 variables
    # Create new variable varNAND, create constrains:
    # not(varNAND) <= var1
    # not(varNAND) <= var2 
    # var1 + var2 <= not(varNAND) + 2 - 1
    # return varNAND
    # if onlyConstrains then only construct constrain var1 + var2 <= 1
    @abc.abstractmethod
    def nand2Var(self, m, _var1, _var2): pass

    # NAND (Alternative denial)
    # Create new variable varNAND, create constrains:
    # not(varNAND) <= var1
    # not(varNAND) <= var2 
    # ...
    # not(varNAND <= varN 
    # var1 + var2 + ... + varN <= not(varNAND) + N - 1
    # return varNAND
    # if onlyConstrains then only construct constrain var1 + var2 + ... + varN <= N -1
    @abc.abstractmethod
    def nandVar(self, m, *_var, onlyConstrains = False): pass
  
    # NOR (Joint Denial) 2 variables
    # Create new variable varNOR, create constrains:
    # var1 <= not(varNOR)
    # var2 <= not(varNOR) 
    # var1 + var2 >= not(varNOR)
    # return varNOR
    # if onlyConstrains then only construct constrain var1 + var2 <= 0
    @abc.abstractmethod
    def nor2Var(self, m, _var1, _var2): pass

    # NOR (Joint Denial)
    # Create new variable varNOR, create constrains:
    # var1 <= not(varNOR)
    # var2 <= not(varNOR) 
    # ...
    # varN <= not(varNOR)
    # var1 + var2 + ... + varN >= not(varNOR)
    # return varNOR
    # if onlyConstrains then only construct constrain var1 + var2 + ... + varN <= 0
    @abc.abstractmethod
    def norVar(self, m, *_var, onlyConstrains = False): pass

    # XOR Exclusive Disjunction 
    # Create new variable varXOR, create constrains:
    # var1 + var2 + varXOR <= 2
    # -var1 - var2 + varXOR <= 0
    # var1 - var2 + varXOR >= 0
    # -var1 + var2 + varXOR >= 0
    # return varXOR
    # if onlyConstrains then only construct constrain var1 + var2 <= 1 and var1 + var2 >= 1 
    @abc.abstractmethod
    def xorVar(self, m, _var1, _var2,  onlyConstrains = False): pass

    # IF Implication
    # Create new variable varIF, create constrains:
    # 1 - var1 <= varIF
    # var2 <= varIF
    # 1 - var1 + var2 >= varIF
    # return varIF
    # if onlyConstrains then only construct constrain var1 <= var2
    @abc.abstractmethod
    def ifVar(self, m, _var1, _var2,  onlyConstrains = False): pass

    # XNOR Equivalence (EQ)
    # Create new variable varEQ, create constrains:
    # var1 + var2 - varEQ <= 1
    # var1 + var2 + varEQ >= 1
    # -var1 + var2 + varEQ <= 1
    # var1 - var2 + varEQ <= 1
    # return varEQ
    # if onlyConstrains then only construct constrain var1 >= var2 and var1 <= var2
    @abc.abstractmethod
    def epqVar(self, m, _var1, _var2,  onlyConstrains = False): pass
    
    # Create constrain var == label, return 1
    @abc.abstractmethod
    def fixedVar(self, m, _var, onlyConstrains = False): pass


import numpy as np
import torch
from torch.nn import Module, GRU, Dropout
from allennlp.modules.seq2seq_encoders import PytorchSeq2SeqWrapper
from ...utils import find_base
import itertools


class WrapperModule(Module):
    # use a wrapper to keep pre-requireds and avoid side-effect of sequencial or other modules
    def __init__(self, module, pres):
        Module.__init__(self)
        self.main_module = module
        for pre in pres:
            from .base import ModuleSensor
            for sensor in pre.find(ModuleSensor, lambda s: not s.output_only):
                self.add_module(sensor.fullname, sensor.module)

    def forward(self, *args, **kwargs):
        return self.main_module(*args, **kwargs)

    def get_output_dim(self):
        return self.main_module.get_output_dim()


class BaseModule(Module):
    def __init__(self, output_dim):
        Module.__init__(self)
        self.output_dim = output_dim
        self.default_device = None

    def get_output_dim(self):
        return self.output_dim


class DropoutRNN(BaseModule):
    def __init__(self, embedding_dim, layers=1, dropout=0.5, bidirectional=True):
        output_dim = embedding_dim
        if bidirectional:
            output_dim *= 2
        BaseModule.__init__(self, output_dim)

        self.embedding_dim = embedding_dim
        self.bidirectional = bidirectional

        self.rnn = PytorchSeq2SeqWrapper(GRU(embedding_dim,
                                             embedding_dim,
                                             num_layers=layers,
                                             batch_first=True,
                                             dropout=dropout,
                                             bidirectional=bidirectional))
        self.dropout = Dropout(dropout)

    def forward(self, x, mask):
        return self.dropout(self.rnn(x, mask))


class Concat(Module):
    def forward(self, *x):
        # TODO: flatten
        return torch.cat(x, dim=-1)


class CartesianProduct(Module):
    def forward(self, x, y):  # (b,l1,f1) x (b,l2,f2) -> (b, l1, l2, f1+f2)
        # TODO: flatten
        xs = x.size()
        ys = y.size()
        assert xs[0] == ys[0]
        # torch cat is not broadcasting, do repeat manually
        xx = x.view(xs[0], xs[1], 1, xs[2]).repeat(1, 1, ys[1], 1)
        yy = y.view(ys[0], 1, ys[1], ys[2]).repeat(1, xs[1], 1, 1)
        return torch.cat([xx, yy], dim=3)


class SelfCartesianProduct(CartesianProduct):
    def forward(self, x):
        return CartesianProduct.forward(self, x, x)


class CartesianProduct3(Module):
    # (b,l1,f1) x (b,l2,f2) -> (b, l1, l2, l3, f1+f2+f3)
    def forward(self, x, y, z):
        # TODO: flatten
        xs = x.size()
        ys = y.size()
        zs = z.size()
        assert xs[0] == ys[0] and ys[0] == zs[0]
        # torch cat is not broadcasting, do repeat manually
        xx = x.view(xs[0], xs[1], 1, 1, xs[2]).repeat(1, 1, ys[1], zs[1], 1)
        yy = y.view(ys[0], 1, ys[1], 1, ys[2]).repeat(1, xs[1], 1, zs[1], 1)
        zz = z.view(ys[0], 1, 1, ys[1], zs[2]).repeat(1, xs[1], ys[1], 1, 1)
        return torch.cat([xx, yy, zz], dim=-1)


class SelfCartesianProduct3(CartesianProduct3):
    def forward(self, x):
        return CartesianProduct3.forward(self, x, x, x)


class JointOldCandidate(Module):
    def forward(self, *tensors):
        #import pdb; pdb.set_trace()
        batch = tensors[0].shape[0]
        num = len(tensors)
        # b, l, c
        logits_sum = tensors[0].clone()
        # n, b, l, c
        addends = [logits.clone() for logits in tensors[1:]]
        for i, logits in enumerate(addends):
            for tensor in addends[i:]:
                # b, ..., 1, l, c
                tensor.unsqueeze_(-3)
            # b, l, ..., l, c = b, ..., l, 1, c + b, ..., 1, l, c 
            logits_sum = logits_sum.unsqueeze_(-2) + logits # this are the logits
        # b, l, ..., l ~ {0, 1}
        candidate = logits_sum.argmax(-1) # find the largest index along logits dim
        # remove self pair
        # b, l, ..., l ~ {0}
        candidate_2 = torch.zeros_like(candidate)[0]
        for indices in itertools.product(*(range(d) for d in candidate.shape[1:])):
            if len(indices) == len(set(indices)):
                # if no same index
                candidate_2[indices] = 1
        # b, l, ..., l ~ {0, 1}
        candidate_2 = candidate_2.expand(batch, *((-1,) * num))
        return candidate * candidate_2


class JointCandidate(Module):
    def forward(self, *tensors):
        #import pdb; pdb.set_trace()
        batch = tensors[0].shape[0]
        num = len(tensors)
        # b, l ~ {0, 1}
        candidate_sum = tensors[0].clone().argmax(-1)
        # n-1, b, l ~ {0, 1}
        addends = [logits.clone().argmax(-1) for logits in tensors[1:]]
        for i, candidate_tensor in enumerate(addends):
            for tensor in addends[i:]:
                # b, ..., 1, l ~ {0, 1}
                tensor.unsqueeze_(-2)
            # b, l, ..., l = b, ..., l, 1 + b, ..., 1, l
            candidate_sum = candidate_sum.unsqueeze_(-1) * candidate_tensor
        # b, l, ..., l ~ {0, 1}
        candidate = candidate_sum
        # remove self pair
        # b, l, ..., l ~ {0}
        candidate_2 = torch.zeros_like(candidate)[0]
        for indices in itertools.product(*(range(d) for d in candidate.shape[1:])):
            if len(indices) == len(set(indices)):
                # if no same index
                candidate_2[indices] = 1
        # b, l, ..., l ~ {0, 1}
        candidate_2 = candidate_2.expand(batch, *((-1,) * num))
        return candidate * candidate_2


class NGram(Module):
    def __init__(self, ngram):
        Module.__init__(self)
        self.ngram = ngram

    def forward(self, x):
        #import pdb; pdb.set_trace()
        shifted = []
        size = x.size() # (b, l, c)
        for i in torch.arange(self.ngram):
            shifted_x = torch.zeros((size[0], size[1]+self.ngram, size[2]), device=x.device)
            shifted_x[:, i:i-self.ngram, :] = x
            shifted.append(shifted_x)
        new_x = torch.cat(shifted, dim=-1)
        offset = int((self.ngram-1) / 2)
        return new_x[:, offset:offset-self.ngram, :]


class PairTokenDistance(BaseModule):
    def __init__(self, emb_num, window):
        BaseModule.__init__(self, emb_num)
        self.emb_num = emb_num # must define emb_num (to have lb and ub) before window
        self.window = window

    @property
    def window(self):
        return self._window

    @window.setter
    def window(self, window):
        self._window = window
        ul = np.floor(window / 2)
        self._base = find_base(ul, self.ub - 1)

    @property
    def base(self):
        return self._base

    @property
    def emb_num(self):
        return self._emb_num

    @emb_num.setter
    def emb_num(self, emb_num):
        self._emb_num = emb_num
        self._lb = -np.floor((emb_num - 1) / 2)
        self._ub = np.ceil((emb_num - 1) / 2)

    @property
    def lb(self):
        return self._lb

    @property
    def ub(self):
        return self._ub

    def forward(self, x):
        batch = len(x)
        length = max(len(xx) for xx in x)
        #(l*2)
        dist = torch.arange(-length + 1, length, device=self.default_device)
        rows = []
        for i in range(length):
            rows.append(dist[i:i + length].view(1, -1))
        #(l, l)
        dist = torch.cat(tuple(reversed(rows)))
        #print(dist)
        sign = dist.sign()
        dist = dist.abs()
        dist = dist.float()
        dist = (dist.log() / np.log(self.base) + 1).floor()
        #print(dist)
        dist[dist < 0] = 0
        dist = dist * sign.to(dtype=dist.dtype, device=dist.device)
        dist[dist < self.lb] = self.lb
        dist[dist > self.ub] = self.ub
        dist = dist - self.lb
        dist = dist.long()
        #print(dist)
        #(n, n)
        eye = torch.eye(self.emb_num, device=dist.device)
        #import pdb; pdb.set_trace()
        #(1, l, l, n)
        dist = eye.index_select(0, dist.view(-1)).view(1, length, length, -1)
        #(b, l, l, n)
        dist = dist.repeat(batch, 1, 1, 1)
        return dist


class PairTokenDependencyRelation(BaseModule):
    def __init__(self):
        BaseModule.__init__(self, 6)

    def forward(self, x):
        #import pdb; pdb.set_trace()
        batch = len(x)
        length = max(len(xx) for xx in x)
        #(b,l,l,f)
        dep = torch.zeros((batch, length, length, self.get_output_dim()), device=self.default_device)
        for i, span in enumerate(x):
            for j, token in enumerate(span):
                # children: immediate syntactic children
                for r_token in token.children:
                    dep[i, j, r_token.i, 0] = 1
                # lefts: leftward immediate children
                for r_token in token.lefts:
                    dep[i, j, r_token.i, 1] = 1
                # rights: rightward immediate children
                for r_token in token.rights:
                    dep[i, j, r_token.i, 2] = 1
                # conjuncts: coordinated tokens
                for r_token in token.conjuncts:
                    dep[i, j, r_token.i, 3] = 1
                # head: syntactic parent, or "governor"
                dep[i, j, token.head.i, 4] = 1
                # ancestors: rightmost token of this toke's syntactic descendants
                for r_token in token.ancestors:
                    dep[i, j, r_token.i, 5] = 1

        return dep


class TripPhraseDistRelation(Module):
    def forward(self, x):
        #import pdb; pdb.set_trace()
        batch, length, feat = x.shape
        device = x.device
        dist = torch.zeros(batch, length, length, length, feat * 2, device=device)

        permutations = itertools.permutations(range(0, length), 3)
        # permutations ensure no repeat index at the same time
        for lm, tr, sp in permutations:
            # (batch, lm:sp, feat)
            lm2sp = x[:, min(lm, sp):max(lm, sp), :]
            # (batch, tr:sp, feat)
            tr2sp = x[:, min(tr, sp):max(tr, sp), :]
            # (batch, feat)
            dist[:, lm, tr, sp, :] = torch.cat((lm2sp.mean(dim=1), tr2sp.mean(dim=1)), dim=1)

        return dist


class LowestCommonAncestor(Module):
    def forward(self, token_lists, features):
        #import pdb; pdb.set_trace()
        batch = features.shape[0]
        length = features.shape[1]
        feat = features.shape[2]

        docs = []
        for token_list in token_lists:
            doc = None
            for token in token_list:
                if doc is None:
                    doc = token.doc
                else:
                    assert doc == token.doc
            docs.append(doc)

        # (b,lx,lx)
        lcas = torch.zeros((batch, length, length), device=features.device, dtype=torch.long)
        for doc, lca in zip(docs, lcas):
            # (l,l) ~ [-1, l-1]
            lca_np = doc.get_lca_matrix()
            # (lx,lx) ~ [0, l]
            lca[:lca_np.shape[0], :lca_np.shape[0]] = torch.as_tensor(lca_np) + 1
        # (b, lx+1, f)
        features = torch.cat((torch.zeros((batch, 1, feat), device=features.device), features), dim=1)
        # (b, lx*lx)
        lcas = lcas.view(batch, length * length)
        lcas_emb = []
        for feature, lca in zip(features, lcas):
            # [(1, lx*lx, f)...]
            lcas_emb.append(feature.index_select(0, lca).view(1, length * length, feat))
        # (b, lx, lx, f)
        lcas_emb = torch.cat(lcas_emb, dim=0).view(batch, length, length, feat)

        return lcas_emb


class PairTokenDependencyDistance(BaseModule):
    def __init__(self, emb_num, window):
        BaseModule.__init__(self, emb_num * 2)

        self.emb_num = emb_num
        self.window = window
        #self.emb_dim = 64
        #self.embedding = torch.nn.Embedding(self.emb_num, self.emb_dim)

    @property
    def window(self):
        return self._window

    @window.setter
    def window(self, window):
        self._window = window
        # leave one for zero-distance, making one-distance another unique (b**0===1 <- b!=0)
        self.base = find_base(window, self.emb_num - 1)

    def forward(self, token_lists):
        #import pdb; pdb.set_trace()
        batch = len(token_lists)
        length = max(len(token_list) for token_list in token_lists)

        docs = []
        for token_list in token_lists:
            doc = None
            for token in token_list:
                if doc is None:
                    doc = token.doc
                else:
                    assert doc == token.doc
            docs.append(doc)

        # (b,lx,lx)
        lcas = torch.zeros((batch, length, length), device=self.default_device, dtype=torch.long)

        for doc, lca in zip(docs, lcas):
            # (l,l) ~ [-1, l-1]
            lca_np = doc.get_lca_matrix()
            # (lx,lx) ~ [-1, l-1]
            lca[:lca_np.shape[0], :lca_np.shape[0]] = torch.as_tensor(lca_np)
        lcas = lcas.float()
        # (b,lx,lx) ~ [0, l-1] U {inf}
        lcas[lcas==-1] = torch.as_tensor(np.inf)
        # (lx,)
        index = torch.arange(length, device=lcas.device, dtype=torch.float)
        # (b,lx,lx) = (b,lx,lx) - (lx,1) ~ [0, l-1] U {inf}
        right_dist = (lcas - index).abs()

        # (b,lx,lx) ~ [0, emb_num-2] U {-inf, inf} + 1 ~ [1, emb_num-1] U {-inf, inf}
        right_dist = (right_dist.log() / np.log(self.base) + 1).floor()
        # (b,lx,lx) ~ [0, emb_num-1]
        right_dist[right_dist < 0] = 0
        right_dist[right_dist > self.emb_num - 1] = self.emb_num - 1
        right_dist = right_dist.long()

        #left_dist = right_dist.transpose(1, 2)

        # (lx, lx)
        eye = torch.eye(self.emb_num, device=lcas.device)
        # (b,lx,lx, emb_num)
        #left_oh = eye.index_select(0, left_dist.view(-1)).view(batch, length, length, -1)
        right_oh = eye.index_select(0, right_dist.view(-1)).view(batch, length, length, -1)
        left_oh = right_oh.transpose(1, 2)

        # (b,lx,lx,emb_dim)
        #left_emb = self.embedding(left_dist)
        # (b,lx,lx,emb_dim)
        #right_emb = self.embedding(right_dist)

        # (b,lx,lx, emb_num*2)
        dist = torch.cat((left_oh, right_oh), dim=-1)
        return dist


class Permute(Module):
    def __init__(self, *dims):
        Module.__init__(self)
        self.dims = dims

    def forward(self, x):
        return x.permute(self.dims)


class Uncap(Module):
    def __init__(self, dim, pre_module=None):
        Module.__init__(self)
        self.dim = dim
        self.pre_module = pre_module

    def forward(self, x, mask):
        if self.pre_module:
            x = self.pre_module(x)

        #import pdb; pdb.set_trace()
        batch, en_len, feat = x.shape
        device = x.device
        # note that mask is generated to match this module already

        # estimate ul
        un_len = en_len ** (1. / self.dim)
        # should always be a natural number if dim is matched as input config
        # could be some slight different due to numerical stability
        un_len = int(np.round(un_len))

        dist = torch.zeros_like(x).reshape((batch, *((un_len,) * self.dim), feat))
        # process each differnt length base on mask
        for dd, xx, mm in zip(dist, x, mask):
            # dd - (ul..., f)
            # xx - (el, f)
            # mm - (ul...,)

            # ()
            ull = mm.sum().float() ** (1. / 3)
            ull = ull.round().int()

            # convert ijk index to 1-d array index
            def encap_index(*indices):
                return sum(i * un_len ** j for j, i in enumerate(reversed(indices)))

            # headfirst solution
            # iter through all and move only the last dim together
            for indices in itertools.product(range(ull), repeat=self.dim-1):
                dd[indices] = xx[encap_index(*indices)]

            # more efficient solution:
            # try torch.unfold or torch.as_strided

        return dist



from setuptools import setup, find_packages
import os

DomiKnowS_version = '0.402'
DomiKnowS_version_env = os.getenv('DomiKnowS_VERSION')
DomiKnowS_VERSION_Postfix = os.getenv('DomiKnowS_VERSION_Postfix')

# Override version if set in environment
if DomiKnowS_version_env is not None:
    DomiKnowS_version = DomiKnowS_version_env[1:]
    if DomiKnowS_VERSION_Postfix is not None:
        DomiKnowS_version = DomiKnowS_version + "." + DomiKnowS_VERSION_Postfix + "0"
    
print("Using DomiKnowS version: " + DomiKnowS_version)
    
setup(
    name='DomiKnowS',
    version=DomiKnowS_version,
    description='A library provides integration between Domain Knowledge and Deep Learning.',

    long_description ='The library allows to specify a problem domain with a conceptual [graph](https://github.com/HLR/DomiKnowS/blob/main/docs/developer/KNOWLEDGE.md#graph) including declarations of edges and nodes, as well as [logical constraints](https://github.com/HLR/DomiKnowS/blob/main/docs/developer/KNOWLEDGE.md#constraints) on the graph concepts and relations. [Neural network](https://github.com/HLR/DomiKnowS/blob/main/docs/developer/MODEL.md#model-declaration) outputs bounded to the graph edges and nodes. The logical constraints are converted to ILP and Gurobi Solver is used for inferring. This adds a relational overlay over elements in a network that relates physical concepts in applications. <br> <br> The example running in Google CoLab environment, presenting the usage of the library is [here](https://colab.research.google.com/drive/1FvdePHv3h3NDSTkBw1VKwAmaZFWuGgTi).',
    long_description_content_type='text/markdown',
    
    url='https://github.com/HLR/DomiKnowS',
    author='Andrzej Uszok',
    author_email='auszok@ihmc.org',

    packages=find_packages(include=['domiknows', 'domiknows.*', 'README.md']),
    
    install_requires=[
       'acls>=1.0.2',
       'Owlready2>=0.30',
       'gurobipy',
       'pandas>=1.1.5',
       'torch>=1.8.1',
       'ordered-set',
       'graphviz',
       'pymongo[tls]',
       'dnspython',
       'scikit-learn',
       'tqdm'
    ],
    
    license='MIT',
    
    classifiers=[
        'Intended Audience :: Developers',

        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
)


import gc
import weakref
from .base import BaseGraphTree


@BaseGraphTree.localize_context
class TrialTree(BaseGraphTree):
    def __init__(self, trial, name=None):
        super().__init__(name=name)
        self._trial = weakref.ref(trial)

    @property
    def trial(self):
        return self._trial()

    def what(self):
        wht = BaseGraphTree.what(self)
        if self.trial:
            wht['trial'] = repr(self.trial)
        return wht


class Trial():
    @classmethod
    def clear(cls):
        TrialTree.clear()
        #gc.collect()  # to release memory right away -- too slow

    @classmethod
    def default(cls):
        default_tree = TrialTree.default()
        if default_tree is not None:
            return default_tree.trial
        return None

    def __init__(self, data=None, obsoleted=None, name=None):
        self.trial_node = TrialTree(self, name=name)
        self.data = data or dict()
        self.obsoleted = obsoleted or set()

    @property
    def sup(self):
        if self.trial_node.sup:
            return self.trial_node.sup.trial
        return None

    def __enter__(self):
        return self.trial_node.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        return self.trial_node.__exit__(exc_type, exc_value, traceback)

    def __getitem__(self, key):
        try:
            return self.data.__getitem__(key)
        except KeyError:
            if key in self.obsoleted:
                raise
            sup = self.sup
            if not sup:
                raise
            return sup.__getitem__(key)

    def __setitem__(self, key, obj):
        self.data.__setitem__(key, obj)

    def __delitem__(self, key):
        try:
            self.data.__delitem__(key)
        except KeyError:
            pass
        self.obsoleted.add(key)

    def __iter__(self):
        yield from iter(self.data)
        if self.sup:
            yield from filter(lambda x: (x not in self.data) and (x not in self.obsoleted), self.sup)

    def __len__(self):
        return len(list(iter(self)))

    def items(self):
        for key in iter(self):
            yield key, self[key]

    def keys(self):
        yield from iter(self)

    def values(self):
        for key in iter(self):
            yield self[key]

    def __repr__(self):
        return repr(dict(self))
